# coding: utf-8
"""Decorators."""
from __future__ import annotations

import time
import functools
import weakref
import inspect

from typing import Callable
from textwrap import dedent


def return_straceback_ifexc(func: Callable):
    """
    Decorator for functions that are supposed to return a string for logging purposes (e.g. str)
    Instead of raising an exception, the decorated function returns a string with the
    traceback so that execution can continue.
    """
    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        try:
            return func(*args, **kwargs)
        except Exception:
            import traceback
            return traceback.format_exc()
    return wrapper


def timeit(method):
    """
    timeit decorator adapted from:
    https://medium.com/pythonhive/python-decorator-to-measure-the-execution-time-of-methods-fa04cb6bb36d
    sets the timing of the routine as an attribute of the class
    """
    def timed(self, *args, **kw):
        ts = time.time()
        result = method(self, *args, **kw)
        te = time.time()

        setattr(self,"time_" + method.__name__, (te - ts) * 1000)
        return result
    return timed


def memoized_method(*lru_args, **lru_kwargs):
    """
    Implements lru_cache for class methods.
    It takes the exact same parameters as lru_cache, and works exactly the same.
    However it never passes self to lru_cache and instead uses a per-instance lru_cache.

    Taken from: https://stackoverflow.com/questions/33672412/python-functools-lru-cache-with-class-methods-release-object


    ... example::

            @memoized_method(maxsize=12, typed=False)
            def method(self, a, b):
                ....
    """
    def decorator(func):
        @functools.wraps(func)
        def wrapped_func(self, *args, **kwargs):
            # We're storing the wrapped method inside the instance. If we had
            # a strong reference to self the instance would never die.
            self_weak = weakref.ref(self)

            @functools.wraps(func)
            @functools.lru_cache(*lru_args, **lru_kwargs)
            def cached_method(*args, **kwargs):
                return func(self_weak(), *args, **kwargs)
            setattr(self, func.__name__, cached_method)
            return cached_method(*args, **kwargs)
        return wrapped_func
    return decorator


def dump_args(func: Callable):
    """
    Decorator to print function call details.

    This includes parameters names and effective values.
    """

    def wrapper(*args, **kwargs):
        func_args = inspect.signature(func).bind(*args, **kwargs).arguments
        func_args_str = ", ".join(
            "{} = {!r}".format(*item) for item in func_args.items()
        )
        print(f"{func.__module__}.{func.__qualname__} ( {func_args_str} )")
        return func(*args, **kwargs)

    return wrapper


class Appender:
    r"""
    A function decorator that appends an addendum to the docstring of the target function.
    This decorator should be robust even if func.__doc__ is None
    (for example, if -OO was passed to the interpreter).
    Usage: construct a docstring.Appender with a string to be joined to
    the original docstring. An optional 'join' parameter may be supplied
    which will be used to join the docstring and addendum. e.g.

    add_copyright = Appender("Copyright (c) 2009", join='\n')

    @add_copyright
    def my_dog(has='fleas'):
        "This docstring will have a copyright below"
        pass

    MG took it from:
    https://github.com/pandas-dev/pandas/blob/3a7f956c30528736beaae5784f509a76d892e229/pandas/util/_decorators.py#L156

    MG: Added dedent and debug args.
    """

    def __init__(self, addendum, join='', indents=0, dedent=True, debug=False):
        if indents > 0:
            self.addendum = indent(addendum, indents=indents)
        else:
            self.addendum = addendum
        self.join = join

        # MG additional stuff
        self.dedent = dedent
        self.debug = debug

    def __call__(self, func):
        func.__doc__ = func.__doc__ if func.__doc__ else ''
        self.addendum = self.addendum if self.addendum else ''

        if self.dedent:
            docitems = [dedent(func.__doc__), dedent(self.addendum)]
        else:
            docitems = [func.__doc__, self.addendum]

        if self.debug:
            print("func.__doc__:\n", func.__doc__)
            print("addendum:\n", self.addendum)

        func.__doc__ = dedent(self.join.join(docitems))

        if self.debug:
            print("new func.__doc__:\n", func.__doc__)

        return func


def indent(text: str, indents=1) -> str:
    if not text or not isinstance(text, str):
        return ''
    jointext = ''.join(['\n'] + ['    '] * indents)
    return jointext.join(text.split('\n'))




"""Context managers"""
from __future__ import annotations

import time
import signal

from contextlib import contextmanager


class Timer:
    """
    Context manager to time code section.

    .. example::

        with Timer("Timing code section"):
            do_stuff()
    """

    def __init__(self, description: str):
        self.description = description

    def __enter__(self):
        self.start = time.time()

    def __exit__(self, exc_type, value, traceback):
        self.end = time.time()
        print(f"{self.description}.\n\tCompleted in {self.end - self.start:.4f} seconds\n")


@contextmanager
def temporary_change_attributes(something, **kwargs):
    """
    https://stackoverflow.com/questions/38531851/how-to-assign-member-variables-temporarily/38532086

    Usage:

        class Something(object):
            def __init__(self, x, y):
                self.x = x
                self.y = y

            def say_hello(self):
                print("hello", self.x, self.y)

        s = Something(1, 2)
        s.say_hello()  # hello 1 2
        with temporary_change_attributes(s, x=4, y=5):
            s.say_hello()  # hello 4 5
        s.say_hello()  # hello 1 2
    """
    previous_values = {k: getattr(something, k) for k in kwargs}
    for k, v in kwargs.items():
        setattr(something, k, v)
    try:
        yield
    finally:
        for k, v in previous_values.items():
            setattr(something, k, v)


class Timeout:
    """
    Taken from https://stackoverflow.com/questions/2281850/timeout-function-if-it-takes-too-long-to-finish/22348885#22348885
    """

    def __init__(self, seconds: int, message: str = 'Timeout'):
        self.seconds = int(seconds)
        if self.seconds <= 0:
            raise ValueError(f"seconds should be > 0 while it is: {self.seconds}")
        self.message = message

    def handle_timeout(self, signum, frame):
        raise TimeoutError(self.message)

    def __enter__(self):
        signal.signal(signal.SIGALRM, self.handle_timeout)
        signal.alarm(self.seconds)

    def __exit__(self, type, value, traceback):
        # If seconds is zero, any pending alarm is canceled.
        signal.alarm(0)



#!/usr/bin/env python
r"""
Equation of state
=================

Flow to compute the equation of state by fitting E(V) at T = 0.
"""
import sys
import os
import abipy.data as abidata
import abipy.abilab as abilab
import abipy.flowtk as flowtk


def build_flow(options):
    # Set working directory (default is the name of the script with '.py' removed and "run_" replaced by "flow_")
    if not options.workdir:
        options.workdir = os.path.basename(sys.argv[0]).replace(".py", "").replace("run_", "flow_")

    # Build GS input file.
    pseudos = abidata.pseudos("Si.GGA_PBE-JTH-paw.xml")
    #silicon = abilab.Structure.zincblende(5.431, ["Si", "Si"], units="ang")
    silicon = abidata.cif_file("si.cif")

    scf_input = abilab.AbinitInput(silicon, pseudos)
    ecut = 12
    scf_input.set_vars(
        ecut=ecut,
        pawecutdg=40,
        nband=6,
        paral_kgb=0,
        iomode=3,
        toldfe=1e-9,
    )

    # K-point sampling (shifted)
    scf_input.set_autokmesh(nksmall=4)

    from abipy.flowtk.gs_works import EosWork
    flow = flowtk.Flow(options.workdir, manager=options.manager)

    # Si is cubic and atomic positions are fixed by symmetry so we
    # use move_atoms=False to compute E(V) with GS-SCF tasks instead of
    # performing a constant-volume optimization of the cell geometry.
    work = EosWork.from_scf_input(scf_input, move_atoms=False, ecutsm=0.5)
    flow.register_work(work)
    flow.allocate(use_smartio=True)

    return flow


# This block generates the thumbnails in the AbiPy gallery.
# You can safely REMOVE this part if you are using this script for production runs.
if os.getenv("READTHEDOCS", False):
    __name__ = None
    import tempfile
    options = flowtk.build_flow_main_parser().parse_args(["-w", tempfile.mkdtemp()])
    build_flow(options).graphviz_imshow()


@flowtk.flow_main
def main(options):
    """
    This is our main function that will be invoked by the script.
    flow_main is a decorator implementing the command line interface.
    Command line args are stored in `options`.
    """
    return build_flow(options)


if __name__ == "__main__":
    sys.exit(main())

############################################################################
#
# Run the script with:
#
#     run_eos.py -s
#
# then use:
#
#    abirun.py flow_eos structures
#
# to get info about the initial and final structures.
#
# .. code-block:: bash
#
#        Lattice parameters:
#                  formula  natom  angle0  angle1  angle2      a      b      c  volume  \
#        w0_t0_in      Si2      2    60.0    60.0    60.0  3.854  3.854  3.854  40.479
#        w0_t0_out     Si2      2    60.0    60.0    60.0  3.854  3.854  3.854  40.479
#        w0_t1_in      Si2      2    60.0    60.0    60.0  3.857  3.857  3.857  40.582
#        w0_t1_out     Si2      2    60.0    60.0    60.0  3.857  3.857  3.857  40.582
#        w0_t2_in      Si2      2    60.0    60.0    60.0  3.861  3.861  3.861  40.684
#        w0_t2_out     Si2      2    60.0    60.0    60.0  3.861  3.861  3.861  40.684
#        w0_t3_in      Si2      2    60.0    60.0    60.0  3.864  3.864  3.864  40.786
#        w0_t3_out     Si2      2    60.0    60.0    60.0  3.864  3.864  3.864  40.786
#        w0_t4_in      Si2      2    60.0    60.0    60.0  3.867  3.867  3.867  40.888
#        w0_t4_out     Si2      2    60.0    60.0    60.0  3.867  3.867  3.867  40.888
#        w0_t5_in      Si2      2    60.0    60.0    60.0  3.870  3.870  3.870  40.991
#        w0_t5_out     Si2      2    60.0    60.0    60.0  3.870  3.870  3.870  40.991
#        w0_t6_in      Si2      2    60.0    60.0    60.0  3.873  3.873  3.873  41.093
#        w0_t6_out     Si2      2    60.0    60.0    60.0  3.873  3.873  3.873  41.093
#        w0_t7_in      Si2      2    60.0    60.0    60.0  3.877  3.877  3.877  41.195
#        w0_t7_out     Si2      2    60.0    60.0    60.0  3.877  3.877  3.877  41.195
#        w0_t8_in      Si2      2    60.0    60.0    60.0  3.880  3.880  3.880  41.297
#        w0_t8_out     Si2      2    60.0    60.0    60.0  3.880  3.880  3.880  41.297
#
#                  abispg_num  P [GPa]  Max|F| eV/ang task_class              status
#        w0_t0_in        None      NaN            NaN    ScfTask  Completed
#        w0_t0_out        227    1.327      9.506e-28    ScfTask  Completed
#        w0_t1_in        None      NaN            NaN    ScfTask  Completed
#        w0_t1_out        227    1.092      1.877e-27    ScfTask  Completed
#        w0_t2_in        None      NaN            NaN    ScfTask  Completed
#        w0_t2_out        227    0.859      4.649e-27    ScfTask  Completed
#        w0_t3_in        None      NaN            NaN    ScfTask  Completed
#        w0_t3_out        227    0.629      5.062e-27    ScfTask  Completed
#        w0_t4_in        None      NaN            NaN    ScfTask  Completed
#        w0_t4_out        227    0.403      5.945e-27    ScfTask  Completed
#        w0_t5_in        None      NaN            NaN    ScfTask  Completed
#        w0_t5_out        227    0.178      6.306e-27    ScfTask  Completed
#        w0_t6_in        None      NaN            NaN    ScfTask  Completed
#        w0_t6_out        227   -0.042      1.738e-27    ScfTask  Completed
#        w0_t7_in        None      NaN            NaN    ScfTask  Completed
#        w0_t7_out        227   -0.259      2.725e-27    ScfTask  Completed
#        w0_t8_in        None      NaN            NaN    ScfTask  Completed
#        w0_t8_out        227   -0.474      6.939e-27    ScfTask  Completed
#
#        Use `--verbose` to print atoms.
#
# Silicon is a cubic materials and the positions are fixed by symmeetry so we only need ScfTask to get E(V).
#
# The EOS results are saved in the JSON file in flow_eos/w0/outdata/eos_data.json.
#
# We can also use the GSR robot to compute and plot the EOS.
#
# Use
#
#    abirun.py flow_eos robot GSR
#
# to build a robot for GSR files and start an ipython shell.
# Then, inside ipython, type:
#
# .. code-block:: ipython
#
#        In [1]: %matplotlib
#        In [2]: robot.gridplot_eos()
#
# to produce
#
# .. image:: https://github.com/abinit/abipy_assets/blob/master/run_eos.png?raw=true
#    :alt: Equation of state of Si obtained with different models.
#



"""
This module provides an API to deal with pseudopotential repositories.
A repo is a collection of versioned pseudopotentials files installed within the same root directory.
A repo has a unique name that encodes the XC functional, relativity type, the kind of pseudopotential and the version.
The default root is $HOME/.abinit/pseudos although it is possible to change it via the ABIPY_PSREPOS_ROOT env variable

Note that all pseudos in a repo share the same XC functional, the type e.g. NC or PAW and the
treatment of relativistic corrections although one might have multiple pseudos for the same element.

Due to this ambiguity, a repo cannot be directly used for running calculations in an automatic fashion
hence the user is supposed to specify both the `repo_name` and the `table_name` when constructing a `PseudoTable`.


Usage:

    from abipy.flowtk.psrepos import get_repo_from_name
    repo = get_repo_from_name("ONCVPSP-PBE-SR-PDv0.4")
    pseudos = repo.get_pseudos("standard")
    print(pseudos)
"""
from __future__ import annotations

import os
import time
import abc
import json
import posixpath
import tempfile
import shutil
import hashlib
import requests

from typing import Optional
from urllib.parse import urlsplit
from monty.termcolor import cprint, colored
from pymatgen.io.abinit.pseudos import Pseudo, PseudoTable
from abipy.tools.decorators import memoized_method
from tqdm import tqdm


# Installation directory.
REPOS_ROOT = os.environ.get("ABIPY_PSREPOS_ROOT",
                            default=os.path.join(os.path.expanduser("~"), ".abinit", "pseudos"))


def get_repos_root() -> str:
    """
    Return the path to the installation directory. Create directory if needed.
    """
    if not os.path.exists(REPOS_ROOT):
        os.mkdir(REPOS_ROOT)

    return REPOS_ROOT


def encode_pseudopath(filepath: str) -> str:
    for repo in _ALL_REPOS:
        if filepath.startswith(repo.dirpath):
            return filepath.replace(repo.dirpath, f"@{repo.name}", 1)
    return filepath


def decode_pseudopath(filepath: str) -> str:
    for repo in _ALL_REPOS:
        start = f"@{repo.name}"
        if filepath.startswith(start):
            return filepath.replace(start, repo.dirpath, 1)
    return filepath


def download_repo_from_url(url: str, save_dirpath: str,
                           chunk_size: int = 2 * 1024**2, verbose: int = 0) -> None:
    """
    Dowload file from url.

    Args:
        url: The url from which the targz is taken.
        save_dirpath: The directory in which the tarball is unpacked.
        chunk_size: Chunk size used for downloading the file.
        verbose: Verbosity level
    """
    path = urlsplit(url).path
    filename = posixpath.basename(path)
    #print(path, filename)

    # stream = True is required by the iter_content below
    with requests.get(url, stream=True) as r:
        #tmp_dir = tempfile.mkdtemp()
        with tempfile.TemporaryDirectory(suffix=None, prefix=None, dir=None) as tmp_dir:
            tmp_filepath = os.path.join(tmp_dir, filename)
            if verbose:
                print("Writing temporary file:", tmp_filepath)

            total_size_in_bytes = int(r.headers.get('content-length', 0))
            progress_bar = tqdm(total=total_size_in_bytes, unit='iB', unit_scale=True)

            with open(tmp_filepath, 'wb') as fd:
                for chunk in r.iter_content(chunk_size=chunk_size):
                    fd.write(chunk)
                    progress_bar.update(len(chunk))

            progress_bar.close()
            if total_size_in_bytes != 0 and progress_bar.n != total_size_in_bytes:
                raise RuntimeError(f"Something went wrong while donwloading url: {url}")

            shutil.unpack_archive(tmp_filepath, extract_dir=tmp_dir)

            dirpaths = [os.path.join(tmp_dir, basen) for basen in os.listdir(tmp_dir) if basen != filename]

            if len(dirpaths) != 1:
                raise RuntimeError(f"Expecting single directory, got {dirpaths}")
            if not os.path.isdir(dirpaths[0]):
                raise RuntimeError(f"Expecting single directory, got {dirpaths}")

            if verbose: print(f"Moving {dirpaths[0]} to {save_dirpath}")
            shutil.move(dirpaths[0], save_dirpath)


def md5_for_filepath(filepath: str) -> str:
    """
    Compute and return the md5 of a file.
    """
    with open(filepath, "rt") as fh:
        text = fh.read()
        m = hashlib.md5(text.encode("utf-8"))
        return m.hexdigest()


def get_repo_from_name(repo_name: str) -> PseudosRepo:
    """
    Return a PseudosRepo from its name ``repo_name``.
    Raises KeyError if ``repo_name`` is not registered.
    """
    for repo in _ALL_REPOS:
        if repo.name == repo_name:
            return repo
    else:
        all_names = [repo.name for repo in _ALL_REPOS]
        raise KeyError(f"Couldn't find {repo_name} in the list of registered repos:\n{all_names}")


def get_installed_repos_and_root(dirpath: Optional[str] = None) -> tuple[list[PseudosRepo], str]:
    """
    Return (all_repos, dirpath)
    """
    dirpath = REPOS_ROOT if not dirpath else dirpath
    dir_basenames = [name for name in os.listdir(dirpath) if os.path.isdir(os.path.join(dirpath, name))]
    dirname2repo = {repo.name: repo for repo in _ALL_REPOS}
    return [dirname2repo[dirname] for dirname in dir_basenames if dirname in dirname2repo], dirpath


class Citation:

    def __init__(self, title: str, doi: str):
        self.title = title
        self.doi = doi

    def __str__(self):
        return f"{self.title}\ndoi:{self.doi}"

    #def __hash__(self) -> int:
    #    return hash(self.doi)

    #def __eq__(self, other):
    #    return self.doi == other.doi


class PseudosRepo(abc.ABC):
    """
    Base abstract class for a Pseudopotential repository that is essentially a collection.
    of pseudopotential files with some metadata and some external files that allow us to
    construct a PseudoTable object.
    """

    #accuracies = ["standard", "stringent"]

    def __init__(self, ps_generator: str, xc_name: str, relativity_type: str, project_name: str,
                 version: str, url: str):
        """
        Args:
            ps_generator: Name of the pseudopotential generator.
            xc_name: XC functional.
            relativity_type: SR for scalar-relativistic or FR for fully relativistic.
            project_name: Name of the project associated to this repository.
            version: Version string.
            url: URL from which the targz will be taken.
        """
        if relativity_type not in {"SR", "FR"}:
            raise ValueError(f"Invalid relativity_type: {relativity_type}")

        self.ps_generator = ps_generator
        self.xc_name = xc_name
        self.version = version
        self.project_name = project_name
        self.relativity_type = relativity_type
        self.url = url

    def to_rowdict(self, verbose: int = 0) -> dict:
        """
        Return dict with metadata, useful to build DataFrames.
        """
        row = dict(
            ps_generator=self.ps_generator,
            ps_type=self.ps_type,
            xc_name=self.xc_name,
            relativity_type=self.relativity_type,
            project_name=self.project_name,
            version=self.version,
            installed=self.is_installed(),
            name=self.name,
        )

        if verbose:
            row.update(url=self.url)

        return row

    def __repr__(self) -> str:
        return self.name

    def __str__(self) -> str:
        lines = [self.name]
        app = lines.append
        return "\n".join(lines)

    def __eq__(self, other):
        return self.name == other.name

    @property
    def isnc(self) -> bool:
        """True if norm-conserving repo."""
        return self.ps_type == "NC"

    @property
    def ispaw(self) -> bool:
        """True if PAW repo."""
        return self.ps_type == "PAW"

    @property
    def table_names(self) -> list[str]:
        """
        List of strings with the name of tables provided by this repository
        """
        if not self.is_installed(): return []
        # TODO: Should read table_names from directory
        table_names = ["standard", "stringent"]
        return table_names

    @property
    def dirpath(self) -> str:
        """Absolute path to the directory with the pseudopotentials."""
        return os.path.join(REPOS_ROOT, self.name)

    def is_installed(self) -> bool:
        """True if the repo is already installed in REPOS_ROOT."""
        return os.path.exists(os.path.join(REPOS_ROOT, self.name))

    def install(self, verbose: int = 0) -> None:
        """
        Install the repository in the standard location relative to the `REPOS_ROOT` directory.
        """
        print(f"Downloading repository from: {self.url} ...")
        print(f"Installing {repr(self)} in: {self.dirpath}")
        start = time.time()
        if not os.path.exists(REPOS_ROOT): os.mkdir(REPOS_ROOT)
        download_repo_from_url(self.url, self.dirpath, verbose=verbose)
        self.validate_checksums(verbose)
        print(f"Installation completed successfully in {time.time() - start:.2f} [s]")

    #####################
    # Abstract interface.
    #####################

    @abc.abstractmethod
    def validate_checksums(self, verbose: int) -> None:
        """Validate md5 checksums after download."""

    @property
    @abc.abstractmethod
    def ps_type(self) -> str:
        """Pseudopotential type e.g. NC or PAW"""

    @property
    @abc.abstractmethod
    def name(self):
        """The name of repository built from the metadata. Must be unique"""

    @abc.abstractmethod
    def get_citations(self) -> list[Citation]:
        """Return list of citations for this repository."""

    @abc.abstractmethod
    def get_pseudos(self, table_name: str) -> PseudoTable:
        """
        Build and return the PseudoTable associated to the given table_name
        """


class OncvpspRepo(PseudosRepo):
    """
    A repo containing ONCVPSP pseudopotentials.
    """

    @classmethod
    def from_github(cls, xc_name: str, relativity_type: str, version: str) -> OncvpspRepo:
        """
        Build an OncvpsRepo from a github repository.
        """
        ps_generator, project_name = "ONCVPSP", "PD"

        if relativity_type == "FR":
            # https://github.com/PseudoDojo/ONCVPSP-PBE-FR-PDv0.4/archive/refs/heads/master.zip
            sub_url = f"{ps_generator}-{xc_name}-FR-{project_name}v{version}"
        elif relativity_type == "SR":
            # https://github.com/PseudoDojo/ONCVPSP-PBE-PDv0.4/archive/refs/heads/master.zip
            sub_url = f"{ps_generator}-{xc_name}-{project_name}v{version}"
        else:
            raise ValueError(f"Invalid relativity_type {relativity_type}")

        url = f"https://github.com/PseudoDojo/{sub_url}/archive/refs/heads/master.zip"
        return cls(ps_generator, xc_name, relativity_type, project_name, version, url)

    @property
    def ps_type(self) -> str:
        return "NC"

    @property
    def name(self) -> str:
        # ONCVPSP-PBEsol-PDv0.4/
        # ONCVPSP-PBE-FR-PDv0.4/
        return f"{self.ps_generator}-{self.xc_name}-{self.relativity_type}-{self.project_name}v{self.version}"

    def get_citations(self) -> list[Citation]:
        return [
            Citation(
                title="The PseudoDojo: Training and grading a 85 element optimized norm-conserving pseudopotential table",
                doi="https://doi.org/10.1016/j.cpc.2018.01.012"),
        ]

    def validate_checksums(self, verbose: int) -> None:
        """
        Compare checksums given in the djson file with the ones computed from file after the donwload.
        """
        print(f"\nValidating md5 checksums of {repr(self)}...")
        djson_paths = [os.path.join(self.dirpath, jfile) for jfile in ("standard.djson", "stringent.djson")]

        seen = set()
        errors = []
        for djson_path in djson_paths:
            with open(djson_path, "rt") as fh:
                djson = json.load(fh)

            for symbol, d in djson["pseudos_metadata"].items():
                bname = d["basename"]
                if bname in seen: continue
                seen.add(bname)
                ref_md5 = d["md5"]
                this_path = os.path.join(self.dirpath, symbol, bname)
                this_md5 = md5_for_filepath(this_path)
                if ref_md5 != this_md5:
                    errors.append(f"Different md5 checksums for {this_path}")
                else:
                    if verbose:
                        print(f"MD5 checksum for {this_path} is OK")

        if errors:
            cprint("Checksum test: FAILED", color="red")
            errstr = "\n".join(errors)
            raise ValueError(f"Checksum test failed for the following pseudos:\n{errstr}\n"
                             f"Data is corrupted. Try to download {repr(self)} again")
        else:
            cprint("Checksum test: OK", color="green")

    @memoized_method()
    def get_pseudos(self, table_name: str) -> PseudoTable:
        """
        Build and return the PseudoTable associated to the given table_name.
        Note that we use a per-instance cache to store the results.
        """
        djson_path = os.path.join(self.dirpath, f"{table_name}.djson")
        pseudos = []
        with open(djson_path, "rt") as fh:
            djson = json.load(fh)
            for symbol, d in djson["pseudos_metadata"].items():
                bname = d["basename"]
                pseudo_path = os.path.join(self.dirpath, symbol, bname)
                #print(f"Reading pseudo from {pseudo_path}")
                # FIXME: Bug if ~ in pseudo_path
                pseudo_path = os.path.expanduser(pseudo_path)
                pseudo = Pseudo.from_file(pseudo_path)
                # Attach a fake dojo_report
                # TODO: This part should be rationalized/rewritten
                hints = d["hints"]
                dojo_report = {"hints": hints}
                pseudo.dojo_report = dojo_report
                #print(f"pseudo.filepath after {pseudo.filepath}")
                pseudos.append(pseudo)

        return PseudoTable(pseudos)


class JthRepo(PseudosRepo):
    """
    A Repo containing JTH PAW pseudos.
    """

    @classmethod
    def from_abinit_website(cls, xc_name: str, relativity_type: str, version: str) -> JthRepo:
        ps_generator, project_name = "ATOMPAW", "JTH"
        # https://www.abinit.org/ATOMICDATA/JTH-LDA-atomicdata.tar.gz
        # ATOMPAW-LDA-JTHv0.4
        url = f"https://www.abinit.org/ATOMICDATA/JTH-{xc_name}-atomicdata.tar.gz"
        return cls(ps_generator, xc_name, relativity_type, project_name, version, url)

    @property
    def ps_type(self) -> str:
        return "PAW"

    @property
    def name(self) -> str:
        # ATOMPAW-LDA-JTHv0.4
        return f"{self.ps_generator}-{self.xc_name}-{self.project_name}v{self.version}"

    def validate_checksums(self, verbose: int) -> None:
        print(f"\nValidating md5 checksums of {repr(self)} ...")
        cprint("WARNING: JTH-PAW repository does not support md5 checksums!!!", color="red")

    @memoized_method()
    def get_pseudos(self, table_name: str) -> PseudoTable:
        """
        Build and return the PseudoPotential table associated to the given table_name
        """
        if table_name != "standard":
            raise ValueError(f"JTH table does not support table_name: {table_name}")

        # Read the list of pseudopotential paths (relative to dirpath)
        txt_path = os.path.join(self.dirpath, f"{table_name}.txt")
        with open(txt_path, "rt") as fh:
            relpaths = fh.readlines()
            relpaths = [l for l in relpaths if l.strip()]

        pseudos = []
        for rpath in relpaths:
            pseudo = Pseudo.from_file(os.path.join(self.dirpath, rpath))
            # TODO: Get hints
            pseudos.append(pseudo)

        return PseudoTable(pseudos)

    def get_citations(self) -> list[Citation]:
        return [
            Citation(
                title="Generation of Projector Augmented-Wave atomic data: A 71 element validated table in the XML format",
                doi="https://www.sciencedirect.com/science/article/abs/pii/S0010465513004359?via%3Dihub"),
        ]


def repos_from_names(repo_names: list[str]) -> list[PseudosRepo]:
    """
    Return list of PP Repos from a list of repo_names
    """
    id2repo = {repo.name: repo for repo in _ALL_REPOS}
    # This will fail if we have received an invalid repo_name
    return [id2repo[repo_name] for repo_name in repo_names]


def repo_from_name(repo_name: str) -> PseudosRepo:
    """
    Return PseudosRepo from its name ``repo_name``.
    """
    id2repo = {repo.name: repo for repo in _ALL_REPOS}
    # This will fail if we have received an invalid repo_name.
    return id2repo[repo_name]


def tabulate_repos(repos: list[PseudosRepo], exclude: Optional[list[str]] = None,
                   with_citations: bool = False, verbose: int = 0) -> str:
    """
    Return string with info on a list of PseudosRepo.

    Args:
        repos: List of PseudosRepo
        exclude: List of keys to exclude
        with_citations: True if citations should be reported.
        verbose: Verbosity level.
    """
    bool2color = {True: "green", False: "red"}

    rows = []
    for i, repo in enumerate(repos):
        d = repo.to_rowdict(verbose=verbose)
        if exclude:
            d = {k: d[k] for k in d if k not in exclude}

        if i == 0:
            headers = list(d.keys())

        if "installed" in d:
            d["installed"] = colored(d["installed"], bool2color[d["installed"]])

        rows.append(list(d.values()))

    from tabulate import tabulate
    s = tabulate(rows, headers)
    if not with_citations: return s

    lines = [s, ""]

    proj2citations = {}
    for repo in repos:
        proj2citations[repo.project_name] = repo.get_citations()

    for projname, citations in proj2citations.items():
        lines.append(f"References for the {projname} project:")
        for i, citation in enumerate(citations):
            lines.append(f"\t- {citation.doi}")

    s = "\n".join(lines)

    return s


#############################################################
# Here we register the repositories and build _ALL_REPOS list
#############################################################

def get_all_registered_repos():
    return _ALL_REPOS[:]


_mk_onc = OncvpspRepo.from_github

_ONCVPSP_REPOS = [
    _mk_onc(xc_name="PBEsol", relativity_type="SR", version="0.4"),
    _mk_onc(xc_name="PBEsol", relativity_type="FR", version="0.4"),
    _mk_onc(xc_name="PBE", relativity_type="SR", version="0.4"),
    #_mk_onc(xc_name="PBE", relativity_type="FR", version="0.4"),  FIXME: checksum fails
    _mk_onc(xc_name="LDA", relativity_type="SR", version="0.4"),
    #_mk_onc(xc_name="LDA", relativity_type="FR", version="0.4"),
]

_mk_jth = JthRepo.from_abinit_website

_PAW_REPOS = [
    _mk_jth(xc_name="LDA", relativity_type="SR", version="1.1"),
    _mk_jth(xc_name="PBE", relativity_type="SR", version="1.1"),
]

_ALL_REPOS = _ONCVPSP_REPOS + _PAW_REPOS

# Make sure repo name is unique.
_repo_names = [_repo.name for _repo in _ALL_REPOS]
if len(set(_repo_names)) != len(_repo_names):
    raise RuntimeError(f"Found duplicated repo_names in ALL_REPOS:\nids: {_repo_names}")


if __name__ == "__main__":
    repo = get_repo_from_name("ONCVPSP-PBE-SR-PDv0.4")
    pseudos = repo.get_pseudos("standard")
    print(pseudos)



# coding: utf-8
"""
Works and Flows for GW calculations with the quartic-scaling implementation.
"""
from __future__ import annotations

import os
import numpy as np

from . import wrappers
from .nodes import Node
from .tasks import NscfTask
from .works import Work


class ScreeningWork(Work):
    """
    This work parallelizes the calculation of the q-points of the screening.
    It also provides the callback `on_all_ok` that calls the mrgscr tool
    to merge all the partial SCR files produced.

    The final SCR file is available in the `outdata` directory of the work.
    To use this SCR file as dependency of other tasks, use the syntax:

        deps={scr_work: "SCR"}

    .. rubric:: Inheritance Diagram
    .. inheritance-diagram:: ScreeningWork
    """

    @classmethod
    def from_nscf_task(cls, nscf_task: NscfTask, scr_input, manager=None) -> ScreeningWork:
        """
        Construct a `ScreeningWork` from a |NscfTask| object.

        Args:
            nscf_task: |NscfTask| object which produced the WFK file with empty states
            scr_input: |AbinitInput| object representing a SCREENING calculation.
            manager: |TaskManager| object.
        """
        if not isinstance(nscf_task, NscfTask):
            raise TypeError("task `%s` does not inherit from NscfTask" % nscf_task)

        new = cls(manager=manager)

        # Call abinit to get the list of q-points for the dielectric matrix
        scr_ibz = scr_input.abiget_scr_ibz()

        # Now we can register the task for the different q-points
        for qpoint in scr_ibz.points:
            new.register_scr_task(scr_input.new_with_vars(nqptdm=1, qptdm=qpoint),
                                  deps={nscf_task: "WFK"})

        return new

    @classmethod
    def from_wfk_filepath(cls, wfk_filepath, scr_input, manager=None) -> ScreeningWork:
        """
        Construct a `ScreeningWork` from a WFK filepath and a screening input.

        WARNING: the parameters reported in `scr_input` e.g. k-mesh, nband, ecut etc
        must be consisten with those used to generate the WFK file.
        No consistencty check is done at this level. You have been warned!

        Args:
            wfk_filepath: Path to the WFK file.
            scr_input: |AbinitInput| object representing a SCREENING calculation.
            manager: |TaskManager| object.
        """
        new = cls(manager=manager)
        wfk_node = Node.as_node(wfk_filepath)

        # Get list of q-points for the dielectric matrix
        scr_ibz = scr_input.abiget_scr_ibz()

        # Now we can register the task for the different q-points
        for qpoint in scr_ibz.points:
            new.register_scr_task(scr_input.new_with_vars(nqptdm=1, qptdm=qpoint),
                                  deps={wfk_node: "WFK"})

        return new

    def merge_scr_files(self, remove_scrfiles=True, verbose=0) -> str:
        """
        This method is called when all the q-points have been computed.
        It runs `mrgscr` in sequential on the local machine to produce
        the final SCR file in the outdir of the `Work`.
        If remove_scrfiles is True, the partial SCR files are removed after the merge.
        """
        scr_files = list(filter(None, [task.outdir.has_abiext("SCR") for task in self]))

        self.history.info("Will call mrgscr to merge %s SCR files:\n" % len(scr_files))
        assert len(scr_files) == len(self)

        mrgscr = wrappers.Mrgscr(manager=self[0].manager, verbose=verbose)
        final_scr = mrgscr.merge_qpoints(self.outdir.path, scr_files, out_prefix="out")

        if remove_scrfiles:
            for scr_file in scr_files:
                try:
                    os.remove(scr_file)
                except IOError:
                    pass

        return final_scr

    def on_all_ok(self): # pragma: no cover
        """
        This method is called when all the q-points have been computed.
        It runs `mrgscr` in sequential on the local machine to produce
        the final SCR file in the outdir of the `Work`.
        """
        final_scr = self.merge_scr_files()
        return self.Results(node=self, returncode=0, message="mrgscr done", final_scr=final_scr)



"""
Pydantic models storing Structure objects and associated metadata

Some of these models are inspired to emmet or we try to use field names and models
that can be easily converted into each other if needed.
"""
from __future__ import annotations

from enum import Enum
from typing import Dict, Any, List
from pydantic import Field
from pymatgen.core.composition import Composition
from pymatgen.core.periodic_table import Element
from abipy.core.structure import Structure
from abipy.htc.base_models import AbipyModel


class CrystalSystem(str, Enum):
    """
    The crystal system of the lattice
    """
    tri = "Triclinic"
    mono = "Monoclinic"
    ortho = "Orthorhombic"
    tet = "Tetragonal"
    trig = "Trigonal"
    hex_ = "Hexagonal"
    cubic = "Cubic"


class StructureData(AbipyModel):
    """
    Store structural info set and symmetry metadata.
    """

    #mpid: str = Field(None, description="MP identifier")
    # Optional or in CustomDict?

    #source_id: str  = Field(..., "Identifier e.g. mp-123")
    #source_type: str = Field(..., "source type e.g. MP for the materials project")
    #meta: Dict[str, Any] = Field(None)

    #hall:

    crystal_system: CrystalSystem = Field(
        None, title="Crystal System", description="The crystal system for this lattice"
    )

    spg_symbol: str = Field(
        None,
        title="Space Group Symbol",
        description="The spacegroup symbol for the lattice",
    )

    spg_number: int = Field(
        None,
        title="Space Group Number",
        description="The spacegroup number for the lattice",
    )

    point_group: str = Field(
        None, title="Point Group Symbol", description="The point group for the lattice"
    )

    symprec: float = Field(
        None,
        title="Symmetry Finding Precision",
        description="The precision given to spglib to determine the symmetry of this lattice",
    )

    angle_tolerance: float = Field(
        None,
        description="Tolerance on the angles",
    )

    spglib_version: str = Field(None, title="SPGLib version")

    # Structure metadata
    nsites: int = Field(None, description="Total number of sites in the structure")

    elements: List[Element] = Field(None, description="List of elements in the material")

    nelements: int = Field(None, title="Number of Elements")

    composition: Composition = Field(None, description="Full composition for the material")

    composition_reduced: Composition = Field(
        None,
        title="Reduced Composition",
        description="Simplified representation of the composition",
    )

    formula_pretty: str = Field(
        None,
        title="Pretty Formula",
        description="Cleaned representation of the formula",
    )

    formula_anonymous: str = Field(
        None,
        title="Anonymous Formula",
        description="Anonymized representation of the formula",
    )

    chemsys: str = Field(
        None,
        title="Chemical System",
        description="dash-delimited string of elements in the material",
    )

    volume: float = Field(
        None,
        title="Volume",
        description="Total volume for this structure in Angstroms^3",
    )

    density: float = Field(None, title="Density", description="Density in grams per cm^3")

    density_atomic: float = Field(
        None,
        title="Packing Density",
        description="The atomic packing density in atoms per cm^3",
    )

    structure: Structure = Field(..., description="Abipy Structure object.")

    custom_data: Dict[str, Any] = Field(None, description="Dictionary containing custom data set by the user.")

    @classmethod
    def from_mpid(cls, material_id: str) -> StructureData:
        """
        Initialize the model from a MaterialsProject id.
        """
        structure = Structure.from_mpid(material_id, final=True)
        return cls.from_structure(structure)

    @classmethod
    def from_structure(cls, structure: Structure) -> StructureData:
        """
        Initialize the model from a Structure.
        """
        from pymatgen.symmetry.analyzer import SpacegroupAnalyzer, spglib

        # 0.1 is the value used in Materials Project. 5.0 is the default value for angles.
        symprec, angle_tolerance = 0.1, 5.0
        sg = SpacegroupAnalyzer(structure, symprec=symprec, angle_tolerance=angle_tolerance)
        #symmetry: Dict[str, Any] = {"symprec": symprec}
        if not sg.get_symmetry_dataset():
            symprec, angle_tolerance = 1e-3, 51
            sg = SpacegroupAnalyzer(structure, symprec=symprec, angle_tolerance=angle_tolerance)

        #symmetry.update(
        symmetry = {
                "spg_symbol": sg.get_space_group_symbol(),
                "spg_number": sg.get_space_group_number(),
                "point_group": sg.get_point_group_symbol(),
                "crystal_system": CrystalSystem(sg.get_crystal_system().title()),
                #"hall": sg.get_hall(),
                "spglib_version": spglib.__version__,
                "symprec": symprec,
                "angle_tolerance": angle_tolerance,
            }
        #)

        comp = structure.composition.remove_charges()
        elsyms = sorted(set([str(e.symbol) for e in comp.elements]))
        #symmetry = SymmetryData.from_structure(structure)

        data = {
            "nsites": structure.num_sites,
            "elements": elsyms,
            "nelements": len(elsyms),
            "composition": comp,
            "composition_reduced": comp.reduced_composition,
            "formula_pretty": comp.reduced_formula,
            "formula_anonymous": comp.anonymized_formula,
            "chemsys": "-".join(elsyms),
            "volume": structure.volume,
            "density": structure.density,
            "density_atomic": structure.volume / structure.num_sites,
            #"symmetry": symmetry,
        }

        structure = Structure.as_structure(structure)  #.copy()
        symmetry.update(data)

        return cls(structure=structure, **symmetry)

    def get_title(self) -> str:
        """Return string with metadata. Useful when creating views"""
        return f"Structure: {self.formula_pretty}, {self.spg_symbol} ({self.spg_number}), " + \
               f"{self.crystal_system}, natom: {self.nsites}"

    #def get_panel_view(self, mng_connector: MongoConnector):
    #    return



# coding: utf-8
"""Pseudopotential Generators."""
from __future__ import annotations

import abc
import os
import tempfile
import collections
import shutil
import time

from typing import Union, Optional # Any,
from shutil import which
from monty.termcolor import cprint
from abipy.flowtk.pseudos import Pseudo
from abipy.ppcodes.oncv_parser import OncvParser

import logging
logger = logging.getLogger(__name__)


# Possible status of the PseudoGenerator.

_STATUS2STR = collections.OrderedDict([
    (1, "Initialized"),    # PseudoGenerator has been initialized
    (2, "Running"),        # PseudoGenerator is running.
    (3, "Done"),           # Calculation done, This does not imply that results are OK
    (4, "Error"),          # PP generator error.
    (5, "Completed"),      # Execution completed successfully.
])


class Status(int):
    """
    An integer representing the status of the 'PseudoGenerator`.
    """
    def __repr__(self) -> str:
        return "<%s: %s, at %s>" % (self.__class__.__name__, str(self), id(self))

    def __str__(self) -> str:
        """String representation."""
        return _STATUS2STR[self]

    @classmethod
    def as_status(cls, obj: Union[Status, str]) -> Status:
        """Convert obj into Status."""
        if isinstance(obj, cls):
            return obj
        else:
            # Assume string
            return cls.from_string(obj)

    @classmethod
    def from_string(cls, s: str) -> Status:
        """Return an instance from its string representation."""
        for num, text in _STATUS2STR.items():
            if text == s:
                return cls(num)
        else:
            raise ValueError(f"Wrong string: `{s}`")


class _PseudoGenerator(metaclass=abc.ABCMeta):
    """
    This object receives a string with the input file and generates a pseudopotential.
    It calls the pp generator in a subprocess to produce the results in a temporary directory.
    It also provides an interface to validate/analyze/plot the results
    produced by the pseudopotential code.

    Concrete classes must:

        1) call super().__init__() in their constructor.
        2) the object should have the input file stored in self.input_str

    Attributes:

        workdir: Working directory (output results are produced in workdir)
        status: Flag defining the status of the ps generator.
        retcode: Return code of the code
        parser: Output parser. None if results are not available because
            the calculations is still running or errors
        pseudo:
            :class:`Pseudo` object. None if not available
    """

    # Possible status
    S_INIT = Status.from_string("Initialized")
    S_RUN = Status.from_string("Running")
    S_DONE = Status.from_string("Done")
    S_ERROR = Status.from_string("Error")
    S_OK = Status.from_string("Completed")

    ALL_STATUS = [
        S_INIT,
        S_RUN,
        S_DONE,
        S_ERROR,
        S_OK,
    ]

    # Basenames for stdin/stdout/stderr.
    stdin_basename: str = "run.in"
    stdout_basename: str = "run.out"
    stderr_basename: str = "run.err"

    def __init__(self, workdir: Optional[str] = None) -> None:
        # Set the initial status.
        self.set_status(self.S_INIT)
        self._parser = None

        if workdir is not None:
            workdir = os.path.abspath(workdir)
            if os.path.exists(workdir):
                raise RuntimeError(f"workdir `{workdir}` already exists")
            self.workdir = workdir

        else:
            # Build a temporary directory
            self.workdir = tempfile.mkdtemp(prefix=self.__class__.__name__)

    def __repr__(self) -> str:
        return "<%s at %s>" % (self.__class__.__name__, self.workdir)

    def __str__(self) -> str:
        return "<%s at %s, status=%s>" % (self.__class__.__name__, self.workdir, self.status)

    @property
    def stdin_path(self) -> str:
        """Absolute path of the standard input."""
        return os.path.join(self.workdir, self.stdin_basename)

    @property
    def stdout_path(self) -> str:
        """Absolute path of the standard output."""
        return os.path.join(self.workdir, self.stdout_basename)

    @property
    def stderr_path(self) -> str:
        """Absolute path of the standard error."""
        return os.path.join(self.workdir, self.stderr_basename)

    @property
    def status(self) -> Status:
        """The status of the job."""
        return self._status

    @property
    def retcode(self) -> Union[int, None]:
        """
        Return code of the subprocess.
        None if not available because e.g. the job has not been started yet.
        """
        try:
            return self._retcode
        except AttributeError:
            return None

    @property
    def parser(self):
        return self._parser

    #@property
    #def pseudo(self) -> Union[Pseudo, None]:
    #    """Pseudo object or None if not available"""
    #    try:
    #        return self._pseudo
    #    except AttributeError:
    #        return None

    @property
    def executable(self) -> str:
        """Name of the executable."""
        return self._executable

    @property
    def input_str(self) -> str:
        """String with the input file."""
        return self._input_str

    def start(self) -> int:
        """"
        Run the calculation in a subprocess (non-blocking interface)
        Return 1 if calculation started, 0 otherwise.
        """
        if self.status >= self.S_RUN:
            return 0

        with open(self.stdin_path, "w") as fh:
            fh.write(self.input_str)

        # Start the calculation in a subprocess and return.
        args = [self.executable, "<", self.stdin_path, ">", self.stdout_path, "2>", self.stderr_path]
        self.cmd_str = " ".join(args)

        from subprocess import Popen, PIPE
        self.process = Popen(self.cmd_str, shell=True, stdout=PIPE, stderr=PIPE, cwd=self.workdir)
        self.set_status(self.S_RUN, info_msg="Start on %s" % time.asctime)

        return 1

    def start_and_wait(self) -> int:
        """
        Run the calculation in a subprocess, wait for it and return exit status.
        """
        self.start()
        retcode = self.wait()
        return retcode

    def poll(self) -> int:
        """
        Check if child process has terminated. Set and return returncode attribute.
        """
        self._retcode = self.process.poll()

        if self._retcode is not None:
            self.set_status(self.S_DONE)

        return self._retcode

    def wait(self) -> int:
        """
        Wait for child process to terminate. Set and return returncode attribute.
        """
        self._retcode = self.process.wait()
        self.set_status(self.S_DONE)

        return self._retcode

    def kill(self) -> None:
        """Kill the child."""
        self.process.kill()
        self.set_status(self.S_ERROR)
        self.errors.append("Process has beed killed by host code.")
        self._retcode = self.process.returncode

    def set_status(self, status, info_msg=None):
        """
        Set the status.

        Args:
            status: Status object or string representation of the status
            info_msg: string with human-readable message used in the case of errors (optional)
        """
        assert status in _STATUS2STR
        self._status = status

        if status == self.S_DONE:
            self.check_status()

        return status

    def get_stdin(self) -> str:
        return self.input_str

    def get_stdout(self) -> str:
        """
        Returns a string with the stdout of the calculation.
        """
        with open(self.stdout_path, "rt") as out:
            return out.read()

    def get_stderr(self) -> str:
        """
        Return string with the stderr of the calculation.
        """
        with open(self.stderr_path, "rt") as err:
            return err.read()

    def rmtree(self) -> int:
        """
        Remove the temporary directory. Return exit status
        """
        try:
            shutil.rmtree(self.workdir)
            return 0
        except Exception:
            return 1

    ### ABC PROTOCOL ###

    #@abc.abstractproperty
    #def parser(self):
    #    return _

    @abc.abstractmethod
    def check_status(self):
        """
        This function checks the status of the task by inspecting the output and the
        error files produced by the application
        """

#import enum
#class CalcType(enum.Enum):
#    nor = "non-relativistic"
#    sr = "scalar-relativistic"
#    fr = "fully-relativistic"



class OncvGenerator(_PseudoGenerator):
    """
    This object receives an input file for oncvpsp, a string
    that defines the type of calculation (scalar-relativistic, ...)
    runs the code in a temporary directory and provides methods
    to validate/analyze/plot the final results.

    Attributes:

        retcode: Retcode of oncvpsp
    """

    @classmethod
    def from_file(cls, path: str, calc_type: str, workdir: Optional[str] = None) -> OncvGenerator:
        """
        Build the object from a file containing the input parameters.
        """
        with open(path, "rt") as fh:
            input_str = fh.read()
            return cls(input_str, calc_type, workdir=workdir)

    def __init__(self, input_str: str, calc_type: str, workdir: Optional[str] = None):
        super().__init__(workdir=workdir)

        self._input_str = input_str
        self.calc_type = calc_type

        self._executable = {
            "non-relativistic": which("oncvpspnr.x"),
            "scalar-relativistic": which("oncvpsp.x"),
            "fully-relativistic": which("oncvpspr.x"),
        }[calc_type]

        if self._executable is None:
            msg = "Cannot find oncvpsp executable in $PATH. Use `export PATH=dir_with_oncvps_executable:$PATH`"
            raise RuntimeError(msg)

    def check_status(self):
        """
        Check the status of the run, set and return self.status attribute.
        """
        if self._status == self.S_OK:
            return self._status

        parser = self._parser = OncvParser(self.stdout_path)

        try:
            parser.scan()
        except parser.Error as exc:
            cprint(str(exc), color="red")
            self._status = self.S_ERROR
            return self._status

        logger.info("run_completed:", parser.run_completed)
        if self.status == self.S_DONE and not parser.run_completed:
            logger.info("Run is not completed!")
            self._status = self.S_ERROR

        if parser.run_completed:
            logger.info("setting status to S_OK")
            self._status = self.S_OK

            psp8_filepath = os.path.join(self.workdir, parser.atsym + ".psp8")

            # Write psp8 file.
            psp8_str = parser.get_psp8_str()
            if psp8_str is not None:
                with open(psp8_filepath, "wt") as fh:
                    fh.write(psp8_str)

            # Add UPF string if present.
            upf_str = parser.get_upf_str()
            if upf_str is not None:
                with open(psp8_filepath.replace(".psp8", ".upf"), "wt") as fh:
                    fh.write(upf_str)

            # Initialize self.pseudo from file.
            pseudo = Pseudo.from_file(psp8_filepath)

            # Add md5 checksum to dojo_report
            if pseudo.has_dojo_report:
                pseudo.dojo_report["md5"] = pseudo.compute_md5()
                pseudo.write_dojo_report(report=pseudo.dojo_report)

        if parser.errors:
            logger.warning("setting status to S_ERROR")
            self._status = self.S_ERROR

        return self._status



# coding: utf-8
"""
mayavi_ toolkit.

WARNING: This code is still under development.
"""
import itertools
import numpy as np

DEFAULT_FIGURE_KWARGS = dict(size=(1024, 768), bgcolor=(1, 1, 1), fgcolor=(0, 0, 0))


def get_fig_mlab(figure=None, **kwargs):  # pragma: no cover
    try:
        from mayavi import mlab
    except ImportError as exc:
        print("mayavi is not installed. Use `conda install mayavi` or `pip install mayavi`")
        raise exc

    # To use the full envisage application
    #mlab.options.backend = "envisage"
    #mlab.options.backend = "test"
    #mlab.options.offscreen = True

    if figure is None:
        # Add defaults
        for k, v in DEFAULT_FIGURE_KWARGS.items():
            if k not in kwargs: kwargs[k] = v
        figure = mlab.figure(**kwargs)

    return figure, mlab


def plot_wigner_seitz(lattice, figure=None, **kwargs):  # pragma: no cover
    """
    Adds the skeleton of the Wigner-Seitz cell of the lattice to a mayavi_ figure

    Args:
        lattice: Reciprocal-space |Lattice| object
        figure: mayavi figure, None to plot on the curretn figure
        kwargs: kwargs passed to the mayavi function ``plot3d``. Color defaults to black
            and line_width to 1.

    Returns: mayavi figure
    """
    figure, mlab = get_fig_mlab(figure=figure)

    if "color" not in kwargs:
        kwargs["color"] = (0, 0, 0)
    if "line_width" not in kwargs:
        kwargs["line_width"] = 1
    if "tube_radius" not in kwargs:
        kwargs["tube_radius"] = None

    bz = lattice.get_wigner_seitz_cell()
    for iface in range(len(bz)):
        for line in itertools.combinations(bz[iface], 2):
            for jface in range(len(bz)):
                if iface < jface and any(np.all(line[0] == x) for x in bz[jface])\
                        and any(np.all(line[1] == x) for x in bz[jface]):
                    #do_plot = True
                    #if in_unit_cell:
                    #    kred0 = lattice.get_fractional_coords(line[0])
                    #    kred1 = lattice.get_fractional_coords(line[1])
                    #    do_plot = np.alltrue((kred0 >= 0) & (kred0 <= 0.5) &
                    #                         (kred1 >= 0) & (kred1 <= 0.5))
                    #    print(kred0, kred1, do_plot)
                    #if not do_plot: continue
                    mlab.plot3d(*zip(line[0], line[1]), figure=figure, **kwargs)

    return figure


def plot_unit_cell(lattice, figure=None, **kwargs):  # pragma: no cover
    """
    Adds the unit cell of the lattice to a mayavi_ figure.

    Args:
        lattice: Lattice object
        figure: mayavi figure, None to plot on the curretn figure
        kwargs: kwargs passed to the mayavi function ``plot3d``. Color defaults to black
            and line_width to 1.

    Returns: mayavi figure
    """
    figure, mlab = get_fig_mlab(figure=figure)

    if "color" not in kwargs:
        kwargs["color"] = (0, 0, 0)
    if "line_width" not in kwargs:
        kwargs["line_width"] = 1
    if "tube_radius" not in kwargs:
        kwargs["tube_radius"] = None

    v = 8 * [None]
    v[0] = lattice.get_cartesian_coords([0.0, 0.0, 0.0])
    v[1] = lattice.get_cartesian_coords([1.0, 0.0, 0.0])
    v[2] = lattice.get_cartesian_coords([1.0, 1.0, 0.0])
    v[3] = lattice.get_cartesian_coords([0.0, 1.0, 0.0])
    v[4] = lattice.get_cartesian_coords([0.0, 1.0, 1.0])
    v[5] = lattice.get_cartesian_coords([1.0, 1.0, 1.0])
    v[6] = lattice.get_cartesian_coords([1.0, 0.0, 1.0])
    v[7] = lattice.get_cartesian_coords([0.0, 0.0, 1.0])

    for i, j in ((0, 1), (1, 2), (2, 3), (0, 3), (3, 4), (4, 5), (5, 6),
                 (6, 7), (7, 4), (0, 7), (1, 6), (2, 5), (3, 4)):
        mlab.plot3d(*zip(v[i], v[j]), figure=figure, **kwargs)

    #mlab.xlabel("x-axis")
    #mlab.ylabel("y-axis")
    #mlab.zlabel("z-axis")

    return figure


def plot_lattice_vectors(lattice, figure=None, **kwargs): # pragma: no cover
    """
    Adds the basis vectors of the lattice provided to a mayavi_ figure.

    Args:
        lattice: |Lattice| object.
        figure: mayavi figure, None if a new figure should be created.
        kwargs: kwargs passed to the mayavi function ``plot3d``. Color defaults to black
            and line_width to 1.

    Returns: mayavi figure
    """
    figure, mlab = get_fig_mlab(figure=figure)

    if "color" not in kwargs:
        kwargs["color"] = (0, 0, 0)
    if "line_width" not in kwargs:
        kwargs["line_width"] = 1
    if "tube_radius" not in kwargs:
        kwargs["tube_radius"] = None

    vertex1 = lattice.get_cartesian_coords([0.0, 0.0, 0.0])
    vertex2 = lattice.get_cartesian_coords([1.0, 0.0, 0.0])
    mlab.plot3d(*zip(vertex1, vertex2), figure=figure, **kwargs)
    vertex2 = lattice.get_cartesian_coords([0.0, 1.0, 0.0])
    mlab.plot3d(*zip(vertex1, vertex2), figure=figure, **kwargs)
    vertex2 = lattice.get_cartesian_coords([0.0, 0.0, 1.0])
    mlab.plot3d(*zip(vertex1, vertex2), figure=figure, **kwargs)

    return figure


def plot_structure(structure, frac_coords=False, to_unit_cell=False, style="points+labels",
                   unit_cell_color=(0, 0, 0), color_scheme="VESTA", figure=None, show=False, **kwargs):  # pragma: no cover
    """
    Plot structure with mayavi.

    Args:
        structure: |Structure| object
        frac_coords:
        to_unit_cell: True if sites should be wrapped into the first unit cell.
        style: "points+labels" to show atoms sites with labels.
        unit_cell_color:
        color_scheme: color scheme for atom types. Allowed values in ("Jmol", "VESTA")
        figure:
        kwargs:

    Returns: mayavi figure
    """
    figure, mlab = get_fig_mlab(figure=figure)

    #if not frac_coords:
    plot_unit_cell(structure.lattice, color=unit_cell_color, figure=figure)
    from pymatgen.analysis.molecule_structure_comparator import CovalentRadius
    from pymatgen.vis.structure_vtk import EL_COLORS

    for site in structure:
        symbol = site.specie.symbol
        color = tuple(i / 255 for i in EL_COLORS[color_scheme][symbol])
        radius = CovalentRadius.radius[symbol]
        if to_unit_cell and hasattr(site, "to_unit_cell"): site = site.to_unit_cell
        x, y, z = site.frac_coords if frac_coords else site.coords

        if "points" in style:
            mlab.points3d(x, y, z, figure=figure, scale_factor=radius,
                          resolution=20, color=color, scale_mode='none', **kwargs)
        if "labels" in style:
            mlab.text3d(x, y, z, symbol, figure=figure, color=(0, 0, 0), scale=0.2)

    if show: mlab.show()
    return figure


def plot_labels(labels, lattice=None, coords_are_cartesian=False, figure=None, **kwargs):  # pragma: no cover
    """
    Adds labels to a mayavi_ figure.

    Args:
        labels: dict containing the label as a key and the coordinates as value.
        lattice: |Lattice| object used to convert from reciprocal to cartesian coordinates
        coords_are_cartesian: Set to True if you are providing.
            coordinates in cartesian coordinates. Defaults to False.
            Requires lattice if False.
        figure: mayavi figure, None to plot on the curretn figure
        kwargs: kwargs passed to the mayavi function `text3d`. Color defaults to blue and size to 25.

    Returns: mayavi figure
    """
    figure, mlab = get_fig_mlab(figure=figure)

    #if "color" not in kwargs:
    #    kwargs["color"] = "b"
    #if "size" not in kwargs:
    #    kwargs["size"] = 25
    #if "width" not in kwargs:
    #    kwargs["width"] = 0.8
    if "scale" not in kwargs:
        kwargs["scale"] = 0.1

    for k, coords in labels.items():
        label = k
        if k.startswith("\\") or k.find("_") != -1:
            label = "$" + k + "$"
        off = 0.01
        if coords_are_cartesian:
            coords = np.array(coords)
        else:
            if lattice is None:
                raise ValueError("coords_are_cartesian False requires the lattice")
            coords = lattice.get_cartesian_coords(coords)
        x, y, z = coords + off
        mlab.text3d(x, y, z, label, figure=figure, **kwargs)

    return figure


class MayaviFieldAnimator(object): # pragma: no cover

    def __init__(self, filepaths):
        self.filepaths = filepaths
        self.num_files = len(filepaths)

    def volume_animate(self):
        from abipy import abilab
        with abilab.abiopen(self.filepaths[0]) as nc:
            nsppol, nspden, nspinor = nc.nsppol, nc.nspden, nc.nspinor
            structure = nc.structure
            datar = nc.field.datar
            # [nspden, nx, ny, nz] array
            nx, ny, nz = datar.shape[1:]
            s = datar[0]
            print(s.dtype, s.shape)

        #cart_coords = np.empty((nx*ny*nz, 3))
        #cnt = 0
        #for i in range(nx):
        #    for j in range(ny):
        #        for k in range(nz):
        #            cart_coords[ctn, :] = (i/nx, j/ny, k/nz)
        #            cnt += 1
        #cart_coords = structure.lattice.get_cartesian_coords(cart_coords)
        # We reorder the points, scalars and vectors so this is as per VTK's
        # requirement of x first, y next and z last.
        #pts = pts.transpose(2, 1, 0, 3).copy()
        #pts.shape = pts.size / 3, 3
        #scalars = scalars.T.copy()
        #vectors = vectors.transpose(2, 1, 0, 3).copy()
        #vectors.shape = vectors.size / 3, 3

        #from tvtk.api import tvtk
        #sgrid = tvtk.StructuredGrid(dimensions=(dims[1], dims[0], dims[2]))
        #sgrid.points = pts
        #s = random.random((dims[0]*dims[1]*dims[2]))
        #sgrid.point_data.scalars = ravel(s.copy())
        #sgrid.point_data.scalars.name = 'scalars'

        figure, mlab = get_fig_mlab(figure=None)
        source = mlab.pipeline.scalar_field(s)
        data_min, data_max = s.min(), s.max()
        print(data_min, data_max)
        #mlab.pipeline.volume(source)
        #                     #vmin=data_min + 0.65 * (data_max - data_min),
        #                     #vmax=data_min + 0.9 * (data_max - data_min))
        #mlab.pipeline.iso_surface(source)
        mlab.pipeline.image_plane_widget(source, plane_orientation='x_axes', slice_index=0)
        mlab.pipeline.image_plane_widget(source, plane_orientation='y_axes', slice_index=0)
        mlab.pipeline.image_plane_widget(source, plane_orientation='z_axes', slice_index=0)

        @mlab.show
        @mlab.animate(delay=1000, ui=True)
        def anim():
            """Animate."""
            t = 1
            while True:
                #vmin, vmax = .1 * np.max(data[t]), .2 * np.max(data[t])
                #print 'animation t = ',tax[t],', max = ',np.max(data[t])
                with abilab.abiopen(self.filepaths[t]) as nc:
                    print("Animation step", t, "from file:", self.filepaths[t])
                    #nsppol, nspden, nspinor = nc.nsppol, nc.nspden, nc.nspinor
                    datar = nc.field.datar
                    # [nspden, nx, ny, nz] array
                    #nx, ny, nz = datar.shape[1:]
                    scalars = datar[0]

                #data_min, data_max = scalars.min(), scalars.max(),
                #mlab.pipeline.volume(source, vmin=data_min + 0.65 * (data_max - data_min),
                #                     vmax=data_min + 0.9 * (data_max - data_min))
                source.mlab_source.scalars = scalars

                t = (t + 1) % self.num_files
                yield

        anim()



"""Integration tests for GW flows."""
from __future__ import annotations

import abipy.data as abidata
import abipy.abilab as abilab
import abipy.flowtk as flowtk

#from abipy.core.testing import has_abinit, has_matplotlib


def make_g0w0_inputs(ngkpt, tvars):
    """
    Input files for the calculation of the GW corrections.

    Returns: gs_input, nscf_input, scr_input, sigma_input
    """
    multi = abilab.MultiDataset(structure=abidata.cif_file("si.cif"),
                                pseudos=abidata.pseudos("14si.pspnc"), ndtset=4)

    # This grid is the most economical, but does not contain the Gamma point.
    scf_kmesh = dict(
        ngkpt=ngkpt,
        shiftk=[0.5, 0.5, 0.5,
                0.5, 0.0, 0.0,
                0.0, 0.5, 0.0,
                0.0, 0.0, 0.5]
    )

    # This grid contains the Gamma point, which is the point at which
    # we will compute the (direct) band gap.
    gw_kmesh = dict(
        ngkpt=ngkpt,
        shiftk=[0.0, 0.0, 0.0,
                0.0, 0.5, 0.5,
                0.5, 0.0, 0.5,
                0.5, 0.5, 0.0]
    )

    # Global variables. gw_kmesh is used in all datasets except DATASET 1.
    ecut = 4

    multi.set_vars(
        ecut=ecut,
        pawecutdg=ecut*2 if multi.ispaw else None,
        istwfk="*1",
        paral_kgb=tvars.paral_kgb,
        gwpara=2,
    )
    multi.set_kmesh(**gw_kmesh)

    # Dataset 1 (GS run)
    multi[0].set_kmesh(**scf_kmesh)
    multi[0].set_vars(tolvrs=1e-6, nband=4)

    # Dataset 2 (NSCF run)
    multi[1].set_vars(iscf=-2,
                      tolwfr=1e-10,
                      nband=10,
                      nbdbuf=2)

    # Dataset3: Calculation of the screening.
    multi[2].set_vars(
        optdriver=3,
        nband=8,
        ecutwfn=ecut,
        symchi=1,
        inclvkb=0,
        ecuteps=2.0,
    )

    # Dataset4: Calculation of the Self-Energy matrix elements (GW corrections)
    kptgw = [
         -2.50000000E-01, -2.50000000E-01,  0.00000000E+00,
         -2.50000000E-01,  2.50000000E-01,  0.00000000E+00,
          5.00000000E-01,  5.00000000E-01,  0.00000000E+00,
         -2.50000000E-01,  5.00000000E-01,  2.50000000E-01,
          5.00000000E-01,  0.00000000E+00,  0.00000000E+00,
          0.00000000E+00,  0.00000000E+00,  0.00000000E+00,
      ]

    multi[3].set_vars(
            optdriver=4,
            nband=10,
            ecutwfn=ecut,
            ecuteps=2.0,
            ecutsigx=2.0,
            symsigma=1,
            #gw_qprange=0,
    )

    bdgw = [4, 5]
    multi[3].set_kptgw(kptgw, bdgw)

    return multi.split_datasets()


def itest_g0w0_flow(fwp, tvars):
    """Test flow for G0W0 calculations."""
    scf, nscf, scr, sig = make_g0w0_inputs(ngkpt=[2, 2, 2], tvars=tvars)

    flow = flowtk.g0w0_flow(fwp.workdir, scf, nscf, scr, sig, manager=fwp.manager)
    # Will remove output files at run-time.
    flow.set_garbage_collector()
    flow.build_and_pickle_dump(abivalidate=True)

    for task in flow[0]:
        task.start_and_wait()

    flow.check_status(show=True)
    assert all(work.finalized for work in flow)
    if not flow.all_ok:
        flow.debug()
        raise RuntimeError()

    scf_task = flow[0][0]
    nscf_task = flow[0][1]
    scr_task = flow[0][2]
    sig_task = flow[0][3]

    # Test garbage_collector
    # The WFK|SCR file should have been removed because we call set_garbage_collector
    assert not scf_task.outdir.has_abiext("WFK")
    assert not nscf_task.outdir.has_abiext("WFK")
    assert not scr_task.outdir.has_abiext("SCR")
    assert not scr_task.outdir.has_abiext("SUS")

    # The sigma task should produce a SIGRES file.
    sigfile = sig_task.outdir.list_filepaths(wildcard="*SIGRES.nc")[0]
    assert sigfile
    with abilab.abiopen(sigfile) as sigres:
        sigres.to_string(verbose=2)
        assert sigres.nsppol == 1

    # Test SigmaTask inspect method
    #if has_matplotlib():
        #sig_task.inspect(show=False)

    # Test get_results for Sigma and Scr
    scr_task.get_results()
    sig_task.get_results()

    # Test SCR.nc file (this is optional)
    if scr_task.scr_path:
        with scr_task.open_scr() as scr:
            scr.to_string(verbose=2)
            assert len(scr.wpts) == 2
            assert scr.nwre == 1 and scr.nwim == 1
            for iq, qpoint in enumerate(scr.qpoints[:2]):
                #print(qpoint)
                qpt, iqcheck = scr.reader.find_qpoint_fileindex(qpoint)
                assert iqcheck == iq
                em1 = scr.get_em1(qpoint)
                #print(em1)

    # TODO Add more tests
    #assert flow.validate_json_schema()


def itest_htc_g0w0(fwp, tvars):
    """Testing G0W0Work."""
    structure = abilab.Structure.from_file(abidata.cif_file("si.cif"))
    pseudos = abidata.pseudos("14si.pspnc")

    flow = flowtk.Flow(fwp.workdir, manager=fwp.manager)

    scf_kppa = 10
    nscf_nband = 10
    #nscf_ngkpt = [4,4,4]
    #nscf_shiftk = [0.0, 0.0, 0.0]
    ecut, ecuteps, ecutsigx = 4, 2, 3
    #scr_nband = 50
    #sigma_nband = 50

    extra_abivars = dict(
        ecut=ecut,
        istwfk="*1",
        paral_kgb=tvars.paral_kgb,
    )

    multi = abilab.g0w0_with_ppmodel_inputs(
        structure, pseudos,
        scf_kppa, nscf_nband, ecuteps, ecutsigx,
        ecut=ecut, pawecutdg=None,
        accuracy="normal", spin_mode="unpolarized", smearing=None,
        #ppmodel="godby", charge=0.0, scf_algorithm=None, inclvkb=2, scr_nband=None,
        #sigma_nband=None, gw_qprange=1):

    )
    multi.set_vars(paral_kgb=tvars.paral_kgb)

    scf_input, nscf_input, scr_input, sigma_input = multi.split_datasets()
    work = flowtk.G0W0Work(scf_input, nscf_input, scr_input, sigma_input)

    flow.register_work(work)
    flow.allocate()
    flow.connect_signals()

    fwp.scheduler.add_flow(flow)
    assert fwp.scheduler.start() == 0
    assert not fwp.scheduler.exceptions
    assert fwp.scheduler.nlaunch == 4

    # The sigma task should produce a SCR file.
    assert len(work[2].outdir.list_filepaths(wildcard="*SCR")) == 1

    flow.show_status()
    if not flow.all_ok:
        flow.debug()
        raise RuntimeError()

    assert all(work.finalized for work in flow)

    #assert flow.validate_json_schema()



# coding: utf-8
"""
Tools to analyze the V1QAVG file produced by the E-PH code (eph_task +15 or -15)
"""
from __future__ import annotations
import numpy as np

from collections import OrderedDict
from monty.string import list_strings, marquee
from monty.functools import lazy_property
from abipy.core.structure import Structure
from abipy.tools.plotting import add_fig_kwargs, get_ax_fig_plt, get_axarray_fig_plt
from abipy.tools.typing import Figure
from abipy.core.mixins import AbinitNcFile, Has_Structure, NotebookWriter
from abipy.core.kpoints import Kpath
from abipy.abio.robots import Robot
from abipy.iotools import ETSF_Reader


def _get_style(reim, what, marker=None, markersize=None, alpha=1.0) -> str:
    lw = 1
    symbol, linestyle, linewidth = {
        "v1scf_avg": (r"v1_{\bf q}", "-", lw),
        "v1scf_abs_avg": (r"|v1_{\bf q}|", "-", lw),
        "v1lr_avg": (r"v1_{\mathrm{lr}}", "--", lw),
        "v1lr_abs_avg": (r"|v1_{\mathrm{lr}}|", "--", lw),
        "v1scfmlr_avg": (r"(v1_{\bf q} - v1_{\mathrm{lr}})", "-.", lw),
        "v1scfmlr_abs_avg": (r"(|v1_{\bf q} - v1_{\mathrm{lr}|})", "-.", lw),
        "v1scf_gsmall": (r"v1_{\bf q}(G)", "-", lw),
        "v1lr_gsmall": (r"v1lr_{\bf q}(G)", "-.", lw),
        "v1scfmlr_gsmall": (r"(v1_{\bf q}(G) - v1_{\mathrm{lr}}(G))", "-.", lw),
    }[what]

    return dict(
        marker=marker,
        markersize=markersize,
        label=r"$\langle \%s %s \rangle$" % ({0: "Re", 1: "Im"}[reim], symbol),
        color={0: "blue", 1: "red"}[reim],
        linestyle=linestyle,
        linewidth=linewidth,
        alpha=alpha,
    )


class V1qAvgFile(AbinitNcFile, Has_Structure, NotebookWriter):
    """
    The V1QAVG.nc file contains the average over the unit cell of the periodic part the DFPT scattering potential.
    This file is produced by the E-PH code by setting eph_task to +15 or -15.
    If eph_task is +15, the input DVDB contains a q-mesh and the potentials are interpolated on a list of q-points
    (usually a q-path) specified by the user. In this case the V1QAVG.nc file also contains an extra array
    with Max_r |W(R, r)|, useful to study the decay of the scattering potentials in R-space.
    If eph_task is -15, the netcdf file contains the average for the q-points found in the DVDB file.
    This option is usually used to visualize the ab-initio potentials and compare then with the model for the LR part.
    """

    def __init__(self, filepath: str):
        super().__init__(filepath)
        self.reader = r = ETSF_Reader(filepath)
        # Read medadata
        self.has_zeff = bool(r.read_value("has_zeff"))
        self.has_dielt = bool(r.read_value("has_dielt"))
        self.has_quadrupoles = bool(r.read_value("has_quadrupoles"))
        self.has_efield = bool(r.read_value("has_efield", default=False))
        self.dvdb_add_lr = r.read_value("dvdb_add_lr")
        self.symv1scf = r.read_value("symv1scf")
        self.interpolated = bool(r.read_value("interpolated"))
        self.qdamp = r.read_value("qdamp")

    @lazy_property
    def structure(self) -> Structure:
        """|Structure| object."""
        return self.reader.read_structure()

    @lazy_property
    def qpoints(self) -> Kpath:
        """List of q-points."""
        frac_coords = self.reader.read_value('qpoints')
        return Kpath(self.structure.reciprocal_lattice, frac_coords, ksampling=None)

    @lazy_property
    def has_maxw(self) -> bool:
        """True if ncfile contains Max_r |W(R, r)|"""
        return "maxw" in self.reader.rootgrp.variables

    def close(self) -> None:
        self.reader.close()

    @lazy_property
    def params(self) -> dict:
        """Dict with parameters that might be subject to convergence studies."""
        return {}

    def __str__(self) -> str:
        return self.to_string()

    def to_string(self, verbose=0) -> str:
        """String representation."""
        lines = []; app = lines.append
        app(marquee("File Info", mark="="))
        app(self.filestat(as_string=True))
        app("")
        app(self.structure.to_string(verbose=verbose, title="Structure"))
        app("")
        app(self.qpoints.to_string(verbose=verbose, title="Q-path"))
        app("")
        app("has_dielt: %s, has_zeff: %s, has_quadrupoles: %s, has_efield: %s" % (
            self.has_dielt, self.has_zeff, self.has_quadrupoles, self.has_efield))
        app("dvdb_add_lr: %s, symv1scf: %s, interpolated: %s, qdamp: %s" % (
            self.dvdb_add_lr, self.symv1scf, self.interpolated, self.qdamp))

        return "\n".join(lines)

    def make_ticks_and_labels(self) -> tuple:
        """
        Find the k-point names in the pymatgen database.
        """
        od = OrderedDict()
        # If the first or the last k-point are not recognized in findname_in_hsym_stars
        # matplotlib won't show the full band structure along the k-path
        # because the labels are not defined. So we have to make sure that
        # the labels for the extrema of the path are always defined.
        od[0] = " "

        for idx, qpoint in enumerate(self.qpoints):
            name = qpoint.name if qpoint.name is not None else self.structure.findname_in_hsym_stars(qpoint)
            if name:
                od[idx] = name
                if qpoint.name is None: qpoint.set_name(name)

        last = len(self.qpoints) - 1
        if last not in od: od[last] = " "

        return list(od.keys()), list(od.values())

    #def xsf_write(self, idir=0, ipert=0, ispden=0):
    #    natom = len(self.structure)
    #    ip = idir + 3 * ipert
    #    r = self.reader
    #    ngfft = r.read_value("ngfft")

    #    from abipy.tools.numtools import transpose_last3dims
    #    from abipy.iotools import xsf

    #    def write(datar, filename):
    #        with open(filename, mode="wt") as fh:
    #            xsf.xsf_write_structure(fh, self.structure)
    #            xsf.xsf_write_data(fh, self.structure, datar, add_replicas=True, cplx_mode="abs")

    #    # nctkarr_t("v1r_interpolated", "dp", "two, nfft, nspden, natom3")
    #    var = r.read_variable("v1r_interpolated")
    #    datar = var[ip, ispden, :, 0] + 1j * var[ip, ispden, :, 1]
    #    # Because we are reading Fortran data (z,y,x) and xsf_write_data expects (..., x, y, z)
    #    #print(datar.shape)
    #    datar = np.reshape(datar, np.flip(ngfft))
    #    datar = transpose_last3dims(datar)
    #    write(datar, "v1r_interp_idir%d_ipert%d.xsf" % (idir, ipert))

    #    var = r.read_variable("v1r_lrmodel")
    #    datar = var[ip, ispden, :, 0] + 1j * var[ip, ispden, :, 1]
    #    datar = np.reshape(datar, np.flip(ngfft))
    #    datar = transpose_last3dims(datar)
    #    write(datar, "v1r_lr_idir%d_ipert%d.xsf" % (idir, ipert))

    @add_fig_kwargs
    def plot(self, what_list="all", ispden=0, fontsize=6, sharey=False, **kwargs) -> Figure:
        """
        Plot

        Args:
            what_list:
            ispden: Spin density component to plot.
            fontsize: fontsize for legends and titles
            sharey: True to share y-axes.

        Return: |matplotlib-Figure|
        """
        #all_varnames = ["v1scf_avg", "v1lr_avg", "v1scfmlr_avg", "v1scfmlr_abs_avg", "v1scf_abs_avg"]
        what_list = list_strings(what_list) if what_list != "all" else ["v1scf_avg", "v1lr_avg"]
        data = {}
        for vname in what_list:
            # Fortran array: nctkarr_t("v1scf_avg", "dp", "two, nspden, three, natom, nqpt")
            data[vname] = self.reader.read_value(vname)

        # Build [natom, 3] grid plot.
        natom = len(self.structure)
        nrows, ncols = natom, 3
        ax_mat, fig, plt = get_axarray_fig_plt(None, nrows=nrows, ncols=ncols,
                                               sharex=True, sharey=sharey, squeeze=False)

        xs = np.arange(len(self.qpoints))
        ticks, labels = self.make_ticks_and_labels()

        for ip, ax in enumerate(ax_mat.ravel()):
            idir = ip % 3
            iat = (ip - idir) // 3
            for reim in (0, 1):
                for vname in what_list:
                    ys = data[vname][:, iat, idir, ispden, reim]
                    ax.plot(xs, ys, **_get_style(reim, vname))

            ax.grid(True)
            if iat == natom - 1: ax.set_xlabel("Q Wave Vector")
            if idir == 0: ax.set_ylabel(r"(Hartree/Bohr)")

            if ticks:
                ax.set_xticks(ticks, minor=False)
                ax.set_xticklabels(labels, fontdict=None, minor=False, size=kwargs.get("qlabel_size", "large"))
                ax.set_xlim(ticks[0], ticks[-1])

            if ip == 0:
                ax.legend(loc="best", fontsize=fontsize, shadow=True)

            site = self.structure[iat]
            s = "%s [%.3f, %.3f, %.3f]" % (site.specie.symbol, site.frac_coords[0], site.frac_coords[1], site.frac_coords[2])
            ax.set_title("idir: %d, iat: %d, %s" % (idir, iat, s), fontsize=fontsize)

        return fig

    @add_fig_kwargs
    def plot_gvec(self, gvec, ispden=0, fontsize=6, sharey=False, **kwargs) -> Figure:
        """
        Plot

        Args:
            gvec: G vector in reduced coordinates.
            ispden: Spin density component to plot.
            fontsize: fontsize for legends and titles
            sharey: True to share y-axes.

        Return: |matplotlib-Figure|
        """
        gsmall = self.reader.read_value("gsmall")
        for ig, g in enumerate(gsmall):
            if np.all(g == gvec): break
        else:
            raise RuntimeError("Cannot find gvec %s in gsmall array" % str(gvec))

        # nctkarr_t("v1scf_gsmall", "dp", "two, ngsmall, nspden, three, natom, nqpt"), &
        # nctkarr_t("v1lr_gsmall", "dp", "two, ngsmall, nspden, three, natom, nqpt"), &
        what_list = ["v1scf_gsmall", "v1lr_gsmall"]
        data = {}
        for vname in what_list:
            ncvar = self.reader.read_variable(vname)
            data[vname] = ncvar[..., ig, :]

        # Build [natom, 3] grid plot.
        natom = len(self.structure)
        nrows, ncols = natom, 3
        ax_mat, fig, plt = get_axarray_fig_plt(None, nrows=nrows, ncols=ncols,
                                               sharex=True, sharey=sharey, squeeze=False)

        xs = np.arange(len(self.qpoints))
        ticks, labels = self.make_ticks_and_labels()

        for ip, ax in enumerate(ax_mat.ravel()):
            idir = ip % 3
            iat = (ip - idir) // 3
            for reim in (0, 1):
                for vname in what_list:
                    ys = data[vname][:, iat, idir, ispden, reim]
                    ax.plot(xs, ys, **_get_style(reim, vname))

                # plot difference.
                ys = (data["v1scf_gsmall"][:, iat, idir, ispden, reim]
                     - data["v1lr_gsmall"][:, iat, idir, ispden, reim]) # * 10
                ax.plot(xs, ys, **_get_style(reim, "v1scfmlr_gsmall"))

            ax.grid(True)
            if iat == natom - 1: ax.set_xlabel("Q Wave Vector")
            if idir == 0: ax.set_ylabel(r"(Hartree/Bohr)")

            if ticks:
                ax.set_xticks(ticks, minor=False)
                ax.set_xticklabels(labels, fontdict=None, minor=False, size=kwargs.get("qlabel_size", "large"))
                ax.set_xlim(ticks[0], ticks[-1])

            if ip == 0:
                ax.legend(loc="best", fontsize=fontsize, shadow=True)

            site = self.structure[iat]
            s = "%s [%.3f, %.3f, %.3f]" % (site.specie.symbol, site.frac_coords[0], site.frac_coords[1], site.frac_coords[2])
            ax.set_title("idir: %d, iat: %d, %s" % (idir, iat, s), fontsize=fontsize)

        fig.suptitle("G = %s" % str(gvec))

        return fig

    @add_fig_kwargs
    def plot_maxw(self, scale="semilogy", ax=None, fontsize=8, **kwargs) -> Figure:
        """
        Plot the decay of max_{r,idir,ipert} |W(R,r,idir,ipert)|

        Args:
            scale: "semilogy", "loglog" or "plot".
            ax: |matplotlib-Axes| or None if a new figure should be created.
            fontsize: fontsize for legends and titles

        Return: |matplotlib-Figure|
        """
        if not self.has_maxw: return None
        ax, fig, plt = get_ax_fig_plt(ax=ax)

        # Fortran array: nctkarr_t("maxw", "dp", "nrpt, natom3")
        rmod = self.reader.read_value("rmod")
        maxw = self.reader.read_value("maxw")
        data = np.max(maxw, axis=0)
        f = {"plot": ax.plot, "semilogy": ax.semilogy, "loglog": ax.loglog}[scale]
        f(rmod, data, marker="o", ls=":", lw=0, **kwargs)

        ax.grid(True)
        ax.set_ylabel(r"$Max_{({\bf{r}}, idir, ipert)} \| W({\bf{r}}, {\bf{R}}, idir, ipert) \|$")
        ax.set_xlabel(r"$\|{\bf{R}}\|$ (Bohr)")

        #if kwargs.pop("with_title", True):
        #    ax.set_title("dvdb_add_lr %d, qdamp: %s, symv1scf: %d" % (self.dvdb_add_lr, self.qdamp, self.symv1scf),
        #                 fontsize=fontsize)
        return fig

    @add_fig_kwargs
    def plot_maxw_perts(self, scale="semilogy", sharey=False, fontsize=8, **kwargs) -> Figure:
        """
        Plot the decay of max_r |W(R,r,idir,ipert)| for the individual atomic perturbations.

        Args:
            scale: "semilogy", "loglog" or "plot".
            sharey: True is y-axes should be shared.
            fontsize: fontsize for legends and titles.

        Return: |matplotlib-Figure|
        """
        if not self.has_maxw: return None
        # Build grid of plots.
        natom = len(self.structure)
        ncols, nrows = (2, natom // 2) if natom % 2 == 0 else (1, natom)

        ax_list, fig, plt = get_axarray_fig_plt(None, nrows=nrows, ncols=ncols,
                                                sharex=True, sharey=sharey, squeeze=False)
        ax_list = ax_list.ravel()

        # Fortran array: nctkarr_t("maxw", "dp", "nrpt, natom3")
        nrpt = self.reader.read_dimvalue("nrpt")
        rmod = self.reader.read_value("rmod")
        maxw = np.reshape(self.reader.read_value("maxw"), (natom, 3, nrpt))

        for iatom, ax in enumerate(ax_list.ravel()):
            site = self.structure[iatom]
            title = "{} [{:.4f} {:.4f} {:.4f}]".format(site.specie.symbol, *site.frac_coords)
            ax.set_title(title, fontsize=fontsize)
            f = {"plot": ax.plot, "semilogy": ax.semilogy, "loglog": ax.loglog}[scale]
            f(rmod, maxw[iatom, 0], marker="o", ls=":", lw=0, label="$L_x$" if iatom == 0 else None)
            f(rmod, maxw[iatom, 1], marker="o", ls=":", lw=0, label="$L_y$" if iatom == 0 else None)
            f(rmod, maxw[iatom, 2], marker="o", ls=":", lw=0, label="$L_z$" if iatom == 0 else None)
            ax.grid(True)
            if iatom == 0:
                ax.set_ylabel(r"$Max_{{\bf{r}}} \| W({\bf{r}}, {\bf{R}}) \|$")
                ax.legend(loc="best", fontsize=fontsize, shadow=True)
            if iatom == len(ax_list) - 1: ax.set_xlabel(r"$\|{\bf{R}}\|$ (Bohr)")

        #fig.suptitle("dvdb_add_lr %d, qdamp: %s, symv1scf: %d" % (self.dvdb_add_lr, self.qdamp, self.symv1scf),
        #             fontsize=fontsize)
        return fig

    def yield_figs(self, **kwargs):  # pragma: no cover
        """
        This function generates a predefined list of matplotlib figures with minimal input from the user.
        """
        title = r"$\langle v1_{\bf q} \rangle \,vs\, \langle v1_{\bfq}^{\mathrm{LR}} \rangle$"
        yield self.plot(title=title, show=False)
        yield self.plot(what_list="v1scfmlr_avg", title=r"$v1_{\bf q} - v1_{\bf q}^{\mathrm{LR}}$", show=False)
        #yield self.plot(what_list=["v1scf_abs_avg", "v1lr_abs_avg"], title=r"ABS", show=False)
        #yield self.plot(what_list="v1scfmlr_abs_avg", title=r"$|v1_{\bf q} - v1_{\bf q}^{\mathrm{LR}|}$", show=False)

        for gvec in [[0, 0, 0], [1, 0, 0], [0, 1, 1], [1, 1, 1], [2, 2, 2]]:
            yield self.plot_gvec(gvec, show=False)

        if self.has_maxw:
            if kwargs.get("verbose", 0) > 0:
                yield self.plot_maxw(show=False)
                yield self.plot_maxw_perts(show=False)
            else:
                print("Use verbose to print the decay of W(r,R)")

    def write_notebook(self, nbpath=None) -> str:
        """
        Write a jupyter notebook to ``nbpath``. If nbpath is None, a temporary file in the current
        working directory is created. Return path to the notebook.
        """
        nbformat, nbv, nb = self.get_nbformat_nbv_nb(title=None)

        nb.cells.extend([
            nbv.new_code_cell("ncfile = abilab.abiopen('%s')" % self.filepath),
            nbv.new_code_cell("print(ncfile)"),
            nbv.new_code_cell("ncfile.plot();"),
        ])

        if self.has_maxw:
            nbv.new_code_cell("ncfile.plot_maxw();"),
            nbv.new_code_cell("ncfile.plot_maxw_perts();"),

        return self._write_nb_nbpath(nb, nbpath)


class V1qAvgRobot(Robot):
    """
    This robot analyzes the results contained in multiple V1qAvgFile.nc files.

    .. rubric:: Inheritance Diagram
    .. inheritance-diagram:: V1qAvgRobot
    """
    EXT = "V1QAVG"

    # Absolute tolerance used to compare q-points.
    atol = 1e-2

    @lazy_property
    def qpoints(self):
        """List of q-points."""
        if len(self) == 1: return self.abifiles[0].qpoints

        if (any(len(ncfile.qpoints) != len(self.abifiles[0].qpoints) for ncfile in self.abifiles)):
            raise RuntimeError("Assuming ncfiles with same number of q-points.\nFound %s" % (
                str([len(ncfile.qpoints) for ncfile in self.abifiles])))

        for abifile in self.abifiles[1:]:
            if np.any(np.abs(abifile.qpoints.frac_coords - self.abifiles[0].qpoints.frac_coords) > self.atol):
                for q1, q2 in zip(self.abifiles[0].qpoints, abifile.qpoints):
                    print("q1:", q1, ", q2:", q2)
                raise RuntimeError("Found different q-points with tolerance: %s!" % self.atol)

        return self.abifiles[0].qpoints

    @add_fig_kwargs
    def plot(self, ispden=0, vname="v1scf_avg", sharey=False, fontsize=8, **kwargs) -> Figure:
        """
        Plot

        Args:
            ispden: Spin density component to plot.
            sharey: True to share y-axes.
            fontsize: fontsize for legends and titles

        Return: |matplotlib-Figure|
        """
        # Caveat: No check is done on the consistency among structures.
        ref_file = self.abifiles[0]
        structure = ref_file.structure

        natom = len(structure)
        nrows, ncols = natom, 3
        ax_mat, fig, plt = get_axarray_fig_plt(None, nrows=nrows, ncols=ncols,
                                               sharex=True, sharey=sharey, squeeze=False)

        xs = np.arange(len(self.qpoints))
        ticks, labels = ref_file.make_ticks_and_labels()

        data_file = {abilabel: abifile.reader.read_value(vname) for abilabel, abifile in self.items()}
        style = dict(marker=".", markersize=2)

        for ip, ax in enumerate(ax_mat.ravel()):
            idir = ip % 3
            iat = (ip - idir) // 3

            for ifile, (abilabel, abifile) in enumerate(self.items()):
                v1scf_avg = data_file[abilabel]
                for reim in (0, 1):
                    style = _get_style(reim, vname)
                    label = "%s %s" % (style.pop("label"), abilabel)
                    style["linestyle"] = {0: "-", 1: "--", 2: "-.", 3: ":"}[ifile]
                    ax.plot(xs, v1scf_avg[:, iat, idir, ispden, reim], label=label if ip == 0 else None, **style)

            if ip == 0:
                ax.legend(loc="best", fontsize=fontsize, shadow=True)

            ax.grid(True)
            if iat == natom - 1: ax.set_xlabel("Q Wave Vector")
            if idir == 0: ax.set_ylabel(r"(Hartree/Bohr)")

            site = ref_file.structure[iat]
            s = "%s [%.3f, %.3f, %.3f]" % (site.specie.symbol, site.frac_coords[0], site.frac_coords[1], site.frac_coords[2])
            ax.set_title("idir: %d, iat: %d, %s" % (idir, iat, s), fontsize=fontsize)

            if ticks:
                ax.set_xticks(ticks, minor=False)
                ax.set_xticklabels(labels, fontdict=None, minor=False, size=kwargs.get("qlabel_size", "large"))
                ax.set_xlim(ticks[0], ticks[-1])

        return fig

    @add_fig_kwargs
    def plot_maxw(self, ax=None, **kwargs) -> Figure:
        """
        Plot

        Args:
            ax: |matplotlib-Axes| or None if a new figure should be created.
            fontsize: fontsize for legends and titles

        Return: |matplotlib-Figure|
        """
        if any(not abifile.has_maxw for abifile in self.abifiles): return None
        ax, fig, plt = get_ax_fig_plt(ax=ax)
        for label, abifile in self.items():
            abifile.plot_maxw(ax=ax, label=label, with_title=False, show=False, **kwargs)

        return fig

    def yield_figs(self, **kwargs): # pragma: no cover
        """
        This function *generates* a predefined list of matplotlib figures with minimal input from the user.
        """
        for vname in ("v1scf_avg", "v1scf_abs_avg"):
            yield self.plot(vname=vname, title=vname, show=False)

        if all(abifile.has_maxw for abifile in self.abifiles):
            yield self.plot_maxw(self, show=False, **kwargs)

    def write_notebook(self, nbpath=None) -> str:
        """
        Write a jupyter notebook to `nbpath`. If nbpath is None, a temporary file in the current
        working directory is created. Return path to the notebook.
        """
        nbformat, nbv, nb = self.get_nbformat_nbv_nb(title=None)

        args = [(l, f.filepath) for l, f in self.items()]
        nb.cells.extend([
            #nbv.new_markdown_cell("# This is a markdown cell"),
            nbv.new_code_cell("robot = abilab.V1qAvgRobot(*%s)\nrobot.trim_paths()\nrobot" % str(args)),
            nbv.new_code_cell("robot.plot();"),
        ])

        return self._write_nb_nbpath(nb, nbpath)


#class V1qAvgPlotter(object):
#
#    def __init__(self, v1q_dfpt_path, v1q_frohl_path, v1q_frohl_quad_path, v1q_froh_quad_efield_path):
#        self.v1q_dfpt = V1qAvgFile.from_file(v1q_dfpt_path)
#        #v1q_frohl = V1qAvgFile.from_file(v1q_frohl_path) if v1q_frohl_path is not None else None
#        #v1q_frohl_quad_path =
#        #v1q_frohl_quad_efield_path =
#
#    #def close(self):
#    #    for k in ("v1q_frohl", "v1q_frohl_quad_path", "v1q_frohl_quad_efield_path"):
#    #        ncfile = getattr(self, k)
#    #        if ncfile is not None: ncfile.close()
#
#    @add_fig_kwargs
#    def plot(self, iatom, idir, ax=None, ispden=0, fontsize=6, **kwargs):
#
#        ax, fig, plt = get_ax_fig_plt(ax=ax)
#
#        return fig



# coding: utf-8
"""
Classes and functions for parsing the ONCVPSP output file and plotting the results.
"""
from __future__ import annotations

import os
import re
import tempfile
import numpy as np
import pandas as pd

from collections import namedtuple, defaultdict
from typing import Union, Any
from dataclasses import dataclass
from monty.functools import lazy_property
from monty.collections import AttrDict, dict2namedtuple
from monty.termcolor import colored
from abipy.core.atom import NlkState, RadialFunction, RadialWaveFunction, l2char
from abipy.ppcodes.base_parser import BaseParser


# Object returned by self._grep
GrepResults = namedtuple("GrepResults", "data, start, stop")

# Used to store ae and pp quantities (e.g wavefunctions) in a single object.
AePsNamedTuple = namedtuple("AePsNamedTuple", "ae, ps")

ConvData = namedtuple("ConvData", "l energies values")

AtanLogDer = namedtuple("AtanLogDer", "l, energies, values")


@dataclass
class AtomicLevel:
    """
    Stores the energy levels of the AE isolated atom.
    """
    nlk: NlkState
    eig: float
    occ: float
    is_valence: bool



class OncvParser(BaseParser):
    """
    Object to read and extract data from the output file of oncvpsp.

    Attributes:
        atsym
        Z
        nc
        nv
        iexc
        psfile

    Example:

        parser = OncvParser(filename).scan()

        # To access data:
        parser.radial_wavefunctions

        # To plot data with matplotlib.
        p = parser.get_plotter()
        p.plot_atanlogder_econv()

    """
    # TODO Improve fully-relativistic case.

    def scan(self, verbose: int = 0) -> OncvParser:
        """
        Scan the output file, set `run_completed` attribute.

        Raises: self.Error if invalid file.
        """
        try:
            self._scan(verbose=verbose)
        except Exception as exc:
            raise self.Error(f"Exception while parsing: {self.filepath}") from exc

        #if not self.run_completed:
        #    cprint("oncvpsp output is not completed. Exiting", "red")
        #    return 1

        #if self.errors:
        #    lines.append(f"# ERRORS ({len(self.errors)})")
        #    lines.extend([colored(s, "red") for s in self.errors])

        return self

    def _scan(self, verbose: int = 0) -> OncvParser:
        if not os.path.exists(self.filepath):
            raise self.Error(f"File {self.filepath} does not exist")

        # Read data and store it in lines
        self.lines = []
        import io
        with io.open(self.filepath, "rt") as fh:
            for i, line in enumerate(fh):
                #print(line)
                line = line.strip()
                self.lines.append(line)

                if verbose and line.startswith("fcfact*="):
                    print(line)

                if line.startswith("DATA FOR PLOTTING"):
                    self.run_completed = True

                # lines that contain the word ERROR but do not seem to indicate an actual teminating error
                acceptable_error_markers = [
                  'run_config: ERROR for fully non-local  PS atom,'
                ]

                if "ERROR" in line:
                    # Example:
                    # test_data: must have fcfact>0.0 for icmod= 1
                    # ERROR: test_data found   1 errors; stopping
                    if line in acceptable_error_markers:
                        self._warnings.append("\n".join(self.lines[i-1:i+1]))
                    else:
                        self._errors.append("\n".join(self.lines[i-1:i+1]))

                if "WARNING" in line:
                    self._warnings.append("\n".join(self.lines[i:i+2]))

                if "GHOST(+)" in line:
                    self._warnings.append(line)
                if "GHOST(-)" in line:
                    self._errors.append(line)

        # Get gendate, calc_type and version
        # scalar-relativistic version 2.1.1, 03/26/2014
        # scalar-relativistic version 3.0.0 10/10/2014
        toks = self.lines[1].replace(",", " ").split()
        self.gendate = toks.pop(-1)
        self.calc_type, self.version = toks[0], toks[2]

        #print(self.version)
        self.major_version, self.minor_version, self.patch_level = tuple(map(int, self.version.split(".")))[:3]

        # Read configuration (not very robust because we assume the user didn't change the template but oh well)

        # Also, handle atom with z > 100 in which atsym and z are not separated by white space e.g.
        #
        # atsym  z   nc   nv     iexc    psfile
        # Rf104.00   10    8       4      both

        header = "# atsym  z    nc    nv    iexc   psfile"
        for i, line in enumerate(self.lines):

            if line.startswith("# atsym"):
                values = self.lines[i + 1].split()
                if len(values) != 6:
                  # Rf104.00   10    8       4      both
                  l = values.pop(0)
                  atmsym, z = l[0:2], l[2:]
                  values.insert(0, z)
                  values.insert(0, atmsym)
                  #print(values)

                keys = header[1:].split()
                # assert len(keys) == len(values)
                # Store values in self.
                for k, v in zip(keys, values):
                    # Convert nc and nv to int.
                    if k in ("nc", "nv", "iexc"): v = int(v)
                    if k in ("z", ): v = float(v)
                    setattr(self, k, v)
                break

        # Parse pseudization options for the local part.
        header = "# lloc, lpopt,  rc(5),   dvloc0"
        self.rc5 = None
        for i, line in enumerate(self.lines):
            if line.startswith(header):
                tokens = self.lines[i + 1].split()
                #print("tokens", tokens)
                self.lloc = int(tokens[0])
                self.lptopt = int(tokens[1])
                self.rc5 = float(tokens[2])
                self.dvloc0 = float(tokens[3])
                break

        if self.rc5 is None:
            raise self.Error(f"Cannot find magic line starting with `{header}` in: {self.filepath}")

        # Parse ATOM and Reference configuration
        # Example:
        """
        #
        #   n    l    f        energy (Ha)
            1    0    2.00    -6.5631993D+01
            2    0    2.00    -5.1265474D+00
            2    1    6.00    -3.5117357D+00
            3    0    2.00    -3.9736459D-01
            3    1    2.00    -1.4998149D-01

        full rel
        in version 4, there no difference between FR and SR
        in version 3, the FR version has:

        #   n    l    f              energy (Ha)
        #   n    l    f        l+1/2             l-1/2
            1    0    2.00    -2.4703720D+03
            2    0    2.00    -4.2419865D+02

        """
        if self.relativistic and self.major_version <= 3:
            header = "#   n    l    f        l+1/2             l-1/2"
        else:
            header = "#   n    l    f        energy (Ha)"

        nc, nv = self.nc, self.nv

        self.atomic_levels = []

        def parse_eigs(str_list):
            """Convert string to float taking into account Fortran numbers in scientific notation."""
            return [float(s.replace("D+", "E+").replace("D-", "E-")) for s in tokens[3:]]

        for i, line in enumerate(self.lines):
            if line.startswith(header):
                # Parse core levels
                beg, is_valence = i + 1, False
                for c in range(nc):
                    tokens = self.lines[beg+c].split()
                    n, l, f = tokens[:3]
                    n, l, f = int(n), int(l), float(f)
                    eigs = parse_eigs(tokens[3:])

                    if not self.relativistic:
                        nlk = NlkState(n=n, l=l, k=None)
                        self.atomic_levels.append(AtomicLevel(nlk, eigs[0], f, is_valence))
                    else:
                        kpa_list = [1, 2] if l != 0 else [1]
                        for ik, kpa in enumerate(kpa_list):
                            nlk = NlkState(n=n, l=l, k=kpa)
                            self.atomic_levels.append(AtomicLevel(nlk, eigs[ik], f, is_valence))

                # Parse valence levels
                beg, is_valence = i + nc + 1, True
                for v in range(nv):
                    #print("lines[beg+v]", self.lines[beg+v])
                    tokens = self.lines[beg+v].split()
                    n, l, f = tokens[:3]
                    n, l, f = int(n), int(l), float(f)
                    eigs = parse_eigs(tokens[3:])

                    if not self.relativistic:
                        nlk = NlkState(n=n, l=l, k=None)
                        self.atomic_levels.append(AtomicLevel(nlk, eigs[0], f, is_valence))
                    else:
                        kpa_list = [1, 2] if l != 0 else [1]
                        for ik, kpa in enumerate(kpa_list):
                            nlk = NlkState(n=n, l=l, k=kpa)
                            self.atomic_levels.append(AtomicLevel(nlk, eigs[ik], f, is_valence))

                break
        else:
            raise self.Error(f"Cannot find header:\n`{header}`\nin output file {self.filepath}")

        return self

    @lazy_property
    def lmax(self) -> int:
        # Read lmax (not very robust because we assume the user didn't change the template but oh well)
        header = "# lmax"
        for i, line in enumerate(self.lines):
            if line.startswith(header):
                return int(self.lines[i+1])
                break
        else:
            raise self.Error(f"Cannot find line with `#lmax` in: {self.filepath}")

    def to_string(self, verbose: int = 0 ) -> str:
        """
        String representation.
        """
        lines = []
        app = lines.append

        if not hasattr(self, "calc_type"):
            app("Object is empty. Call scan method to analyze output file")
            return "\n".join(lines)

        if not self.run_completed:
            app("completed: %s" % self.run_completed)
            return "\n".join(lines)

        app(f"relativity: {self.calc_type}, oncvpsp version: {self.version}, date: {self.gendate}\n")

        df = self.get_atomic_levels_df()
        app("# Atomic levels:")
        app(str(df) + 2 * "\n")
        app("# Peaks of radial wavefunctions:")
        df = self.get_peaks_df()
        app(str(df) + 2 * "\n")

        from pprint import pformat
        app("# Results:\n")
        app(pformat(self.get_results()) + 2*"\n")

        if self.warnings:
            lines.append(f"# WARNINGS ({len(self.warnings)})")
            lines.extend([colored(s, "magenta") for s in self.warnings])

        if self.errors:
            lines.append(f"# ERRORS ({len(self.errors)})")
            lines.extend([colored(s, "red") for s in self.errors])

        return "\n".join(lines)

    def __str__(self) -> str:
        return self.to_string()

    @property
    def relativistic(self) -> bool:
        """True if fully-relativistic calculation."""
        return self.calc_type in ("fully-relativistic", "relativistic")

    @lazy_property
    def rc_l(self) -> dict[int, float]:
        """
        Core radii as a function of l extracted from the output file.
        """
        rc_l = {}
        header = "#   l,   rc,"
        for i, line in enumerate(self.lines):
            if line.startswith(header):
                beg = i + 1
                nxt = 0
                while True:
                    ln = self.lines[beg + nxt]
                    if ln.startswith("#"): break
                    tokens = ln.split()
                    #print("line:", ln, "\ntokens", tokens)
                    l, rc = int(tokens[0]), float(tokens[1])
                    rc_l[l] = rc
                    nxt += 1

        if not rc_l:
            raise self.Error(f"Cannot find magic line starting with `{header}` in: {self.filepath}")

        return rc_l

    @lazy_property
    def kinerr_nlk(self) -> dict[NlkState, namedtuple]:
        """
        Dictionary with the error on the kinetic energy indexed by nlk.
        """

        # In relativistic mode we write data inside the following loops:

        #do l1=1,lmax+1
        #   ll=l1-1
        #   if(ll==0) then
        #    mkap=1
        #   else
        #    mkap=2
        #   end if
        #   do ikap=1,mkap
        #       if(ikap==1) then
        #         kap=-(ll+1)
        #       else
        #         kap=  ll
        #       end if

        kinerr_nlk = {}

        if self.major_version > 3:
            # Calculating optimized projector #   1
            #
            #  for l=   0

            re_start = re.compile(r"^Calculating optimized projector #\s+(?P<iproj>\d+)")

        else:
            # Calculating first optimized projector for l=   0
            re_start = re.compile(r"^Calculating (?P<iproj>(first|second)) optimized projector for l=\s+(?P<l>\d+)")
            # TODO: In FR mode, we have
            #Calculating first optimized projector for l=   0
            #Calculating second optimized projector for l=   0

        nlk = None
        iproj_l_seen = set()

        for i, line in enumerate(self.lines):
            m = re_start.match(line)
            if m:
                # Extract iproj and l.
                if self.major_version > 3:
                    # for l=   0
                    iproj = int(m.group("iproj"))
                    l = int(self.lines[i+2].split("=")[-1].strip())
                else:
                    iproj = m.group("iproj")
                    iproj = {"first": 0, "second": 1}[iproj]
                    l = int(m.group("l"))

                k = None
                if self.relativistic:
                    k = 1
                    if (iproj, l) in iproj_l_seen: k= 2
                    iproj_l_seen.add((iproj, l))

                # Use n index to store iprj index.
                nlk = NlkState(n=iproj, l=l, k=k)
                #print("nlk:", nlk)
                continue

            # Now parse the following section associated to nlk

            #Energy error per electron        Cutoff
            #     Ha          eV             Ha
            #     0.01000     0.27211       27.01
            #     0.00100     0.02721       52.82
            #     0.00010     0.00272       66.22
            #     0.00001     0.00027       75.37

            if line.startswith("Energy error per electron        Cutoff"):
                values_ha, ecuts = [], []
                for j in range(4):
                    tokens = self.lines[i+2+j].split()
                    #print("tokens:", tokens)
                    if not tokens: break
                    err_ha, err_ev, ecut = map(float, tokens)
                    values_ha.append(err_ha)
                    ecuts.append(ecut)

                if nlk is None:
                    raise self.Error("Cannot find nlk quantum numbers")

                self._check_nlk_key(nlk, kinerr_nlk, "kinerr_nlk")

                kinerr_nlk[nlk] = dict2namedtuple(ecuts=ecuts, values_ha=values_ha)

        if not kinerr_nlk:
            raise self.Error(f"Cannot parse convergence profile in: {self.filepath}")

        return kinerr_nlk

    @staticmethod
    def _check_nlk_key(nlk, d, dict_name) -> None:

        if nlk in d:
            ks = "\n\t".join(str(k) for k in d)
            raise RuntimeError(f"nlk state `{nlk}` is already in {dict_name}:\nKeys:\n\t{ks}")

    @lazy_property
    def potentials(self) -> dict[int, RadialFunction]:
        """
        Dict with radial functions with the non-local and local potentials indexed by l.
        l = -1 corresponds to the local part (if present).
        """
        #radii, charge, pseudopotentials (ll=0, 1, lmax)
        #!p   0.0099448   4.7237412  -7.4449470 -14.6551019
        vl_data = self._grep("!p").data
        lmax = len(vl_data[0]) - 3
        assert lmax == self.lmax

        # From 0 up to lmax
        ionpots_l = {}
        for l in range(lmax + 1):
            ionpots_l[l] = RadialFunction("Ion Pseudopotential, l=%d" % l, vl_data[:, 0], vl_data[:, 2+l])

        # Local part is stored with l == -1 if lloc=4, not present if lloc = l
        vloc = self._grep("!L").data
        if vloc is not None:
            ionpots_l[-1] = RadialFunction("Local part, l=%d" % -1, vloc[:, 0], vloc[:, 1])

        return ionpots_l

    @lazy_property
    def densities(self) -> dict[str, RadialFunction]:
        """
        Dictionary with charge densities on the radial mesh.
        """
        # radii, charge, core charge, model core charge
        # !r   0.0100642   4.7238866  53.4149287   0.0000000
        rho_data = self._grep("!r").data

        return dict(
            rhoV=RadialFunction("Valence charge", rho_data[:, 0], rho_data[:, 1]),
            rhoC=RadialFunction("Core charge", rho_data[:, 0], rho_data[:, 2]),
            rhoM=RadialFunction("Model charge", rho_data[:, 0], rho_data[:, 3])
        )

    @lazy_property
    def radial_wfs(self) -> AePsNamedTuple:
        """
        Read and set the radial wavefunctions for the bound states.

        Usage:

            ae_wfs, ps_wfs = self.radial_wfs.ae, self.radial_wfs.ps

            for nlk, ae_wf in ae_wfs.items():
                ps_wf, l, k = ps_wfs[nlk], nlk.l, nlk.k
        """
        return self._get_radial_wavefunctions(what="bound_states")

    @property
    def has_scattering_wfs(self) -> bool:
        """
        True if pp generation included scattering states.
        """
        return bool(self.scattering_wfs.ae)

    @lazy_property
    def scattering_wfs(self) -> AePsNamedTuple:
        """
        Read and set the scattering wavefunctions.
        """
        return self._get_radial_wavefunctions(what="scattering_states")

    def _get_radial_wavefunctions(self, what: str) -> AePsNamedTuple:
        # For scalar-relativistic bound states, we have
        #
        #    n= 1,  l= 0, all-electron wave function, pseudo w-f
        #
        #    &     0    0.009945   -0.092997    0.015273

        # For fully-relativistic bound states, we have:
        #
        #    n= 1,  l= 0  kap=-1, all-electron wave function, pseudo w-f
        #
        #    &     0    0.009955    0.066338    0.000979

        # For the scattering states (scalar and relativistic case)
        #
        # scattering, iprj= 2,  l= 1, all-electron wave function, pseudo w-f
        #
        # scattering, iprj= 2,  l= 1, kap= 1, all-electron wave function, pseudo w-f

        ae_waves, ps_waves = {}, {}

        #l_to_nlist = defaultdict(list)
        #for level in self.atomic_levels:
        #    if not level.is_valence: continue
        #    l_to_nlist[level.nlk.l].append(level.nlk.n)

        beg = 0
        while True:
            g = self._grep("&", beg=beg)
            if g.data is None: break
            beg = g.stop + 1

            # Get header two lines above.
            header = self.lines[g.start - 2]

            if what == "bound_states":
                if header.startswith("scattering,"):
                    continue
            elif what == "scattering_states":
                if not header.startswith("scattering,"):
                    continue
                header = header.replace("scattering,", "")
            else:
                raise ValueError(f"Invalid value of {what=}")
            #print("header:", header)

            if not self.relativistic:
                # n= 1,  l= 0, all-electron wave function, pseudo w-f
                n, l = header.split(",")[0:2]
                n = int(n.split("=")[1])
                l = int(l.split("=")[1])
                # TODO
                #if what == "bound_states" and l_to_nlist[l]:
                #    print(f"for {l=} {l_to_nlist[l]=}")
                #    n = l_to_nlist[l].pop(0)
                kap = None
            else:
                # n= 1,  l= 0,  kap=-1, all-electron wave function, pseudo w-f
                if self.major_version <= 2: header = header.replace("kap=", ", kap=")
                n, l, kap = header.split(",")[0:3]
                n = int(n.split("=")[1])
                l = int(l.split("=")[1])
                kap = int(kap.split("=")[1])

            nlk = NlkState.from_nlkap(n=n, l=l, kap=kap)
            #print("Got nlk state:", nlk)

            rmesh = g.data[:, 1]
            ae_wf = g.data[:, 2]
            ps_wf = g.data[:, 3]

            self._check_nlk_key(nlk, ae_waves, "ae_waves")

            ae_waves[nlk] = RadialWaveFunction(nlk, str(nlk), rmesh, ae_wf)
            ps_waves[nlk] = RadialWaveFunction(nlk, str(nlk), rmesh, ps_wf)

        return AePsNamedTuple(ae=ae_waves, ps=ps_waves)

    @lazy_property
    def projectors(self) -> dict[NlkState, RadialFunction]:
        """
        Dict with projector wave functions indexed by nlk.
        """
        #
        #@     0    0.009945    0.015274   -0.009284
        beg = 0
        magic = "@"
        if self.major_version > 3: magic = "!J"

        # if(ikap==1) then
        #   write(6,'(a,i6,6(f12.6,1x))') '!J',-ll,rr(ii), &
        #        (vkb(ii,jj,l1,ikap),jj=1,nproj(l1))
        # else
        #   write(6,'(a,i6,6(f12.6,1x))') '!J',ll,rr(ii), &
        #        (vkb(ii,jj,l1,ikap),jj=1,nproj(l1))

        projectors_nlk = {}
        while True:
            g = self._grep(magic, beg=beg)
            if g.data is None: break
            beg = g.stop + 1

            rmesh = g.data[:, 1]
            l = int(g.data[0, 0])

            k = None
            if self.relativistic:
                k = 2
                if l <= 0: k = 1

            for n in range(len(g.data[0]) - 2):
                nlk = NlkState(n=n + 1, l=abs(l), k=k)
                #print("Got projector with: %s" % str(nlk))

                if nlk in projectors_nlk:
                    raise self.Error("nlk state `{nlk}` is already in projectors_nlk")

                projectors_nlk[nlk] = RadialWaveFunction(nlk, str(nlk), rmesh, g.data[:, n + 2])

        return projectors_nlk

    @lazy_property
    def atan_logders(self) -> AePsNamedTuple:
        """
        Atan of the log derivatives for different l-values.
        """
        #log derivativve data for plotting, l= 0
        #atan(r * ((d psi(r)/dr)/psi(r))), r=  1.60
        #l, energy, all-electron, pseudopotential
        #
        #!      0    2.000000    0.706765    0.703758
        ae_atan_logder_l, ps_atan_logder_l = {}, {}

        lstop = self.lmax + 1
        if self.major_version > 3:
            lstop = min(self.lmax + 2, 4)

        if not self.relativistic:
            l_list = list(range(lstop))
        else:
            # Order with (l, -l) for plotting purposes.
            l_list = []
            for l in range(lstop):
                if l != 0:
                    l_list.extend((l, -l))
                else:
                    l_list.append(0)

        for l in l_list:
            tag = "!      %d" % l if l >= 0 else "!     %d" % l
            data = self._grep(tag=tag).data
            if data is None:
                raise self.Error(f"Cannot find logder for l: {l}")
            assert l == int(data[0, 0])

            ae_atan_logder_l[l] = AtanLogDer(l=l, energies=data[:, 1], values=data[:, 2])
            ps_atan_logder_l[l] = AtanLogDer(l=l, energies=data[:, 1], values=data[:, 3])

        return AePsNamedTuple(ae=ae_atan_logder_l, ps=ps_atan_logder_l)

    @lazy_property
    def kene_vs_ecut(self) -> dict[int, ConvData]:
        """
        Dict with the convergence of the kinetic energy versus ecut for different l-values.
        """
        #convergence profiles, (ll=0,lmax)
        #!C     0    5.019345    0.010000
        #...
        #!C     1   19.469226    0.010000
        # TODO: This does not take into account scattering states or n > 1
        conv_l = {}

        for l in range(self.lmax + 1):
            data = self._grep(tag="!C     %d" % l).data
            conv_l[l] = ConvData(l=l, energies=data[:, 1], values=data[:, 2])

        return conv_l

    @lazy_property
    def hints(self) -> dict:
        """
        Hints for the cutoff energy as provided by oncvpsp.
        """
        # Extract the hints
        hints = 3 * [-np.inf]
        for i in range(3):
            for l in range(self.lmax + 1):
                hints[i] = max(hints[i], self.kene_vs_ecut[l].energies[-i-1])
        hints.reverse()

        # Truncate to the nearest int
        hints = [np.rint(h) for h in hints]
        # print("hints:", hints)

        return dict(
            low={"ecut": hints[0], "pawecutdg": hints[0]},
            normal={"ecut": hints[1], "pawecutdg": hints[1]},
            high={"ecut": hints[2], "pawecutdg": hints[2]}
        )

    def get_results(self) -> AttrDict:
        """
        Return the most important results extracted from the output file.
        """
        # Init return values
        #d = AttrDict(
        #    max_ecut=None,
        #    max_atan_logder_l1err=None,
        #    max_psexc_abserr=None,
        #    herm_err=None,
        #    nwarns=len(self.warnings)
        #    nerrs=len(self.errors)
        #)

        # Get the max ecut estimated by oncvpsp.
        # TODO: Should take into account scattering states.
        max_ecut = max(self.kene_vs_ecut[l].energies[-1] for l in self.kene_vs_ecut)

        # Compute the l1 error in atag(logder) between AE and PS
        from scipy.integrate import cumtrapz
        max_l1err = 0.0
        for l in self.atan_logders.ae:
            f1, f2 = self.atan_logders.ae[l], self.atan_logders.ps[l]

            abs_diff = np.abs(f1.values - f2.values)
            integ = cumtrapz(abs_diff, x=f1.energies) / (f1.energies[-1] - f1.energies[0])
            max_l1err = max(max_l1err, integ[-1])

        # Read Hermiticity error and compute the max value of PSP excitation error=
        # Hermiticity error    4.8392D-05
        # PSP excitation error=  1.56D-10
        herm_tag, pspexc_tag = "Hermiticity error", "PSP excitation error="
        herm_err, max_psexc_abserr = None, -np.inf

        for line in self.lines:
            i = line.find(herm_tag)
            if i != -1:
                herm_err = float(line.split()[-1].replace("D", "E"))

            i = line.find(pspexc_tag)
            if i != -1:
                max_psexc_abserr = max(max_psexc_abserr, abs(float(line.split()[-1].replace("D", "E"))))

        return AttrDict(
            max_ecut=max_ecut,
            max_atan_logder_l1err=max_l1err,
            max_psexc_abserr=max_psexc_abserr,
            herm_err=herm_err,
            nwarns=len(self.warnings),
            nerrs=len(self.errors),
        )

    def find_string(self, s: str) -> int:
        """
        Returns the index of the first line containing string s.
        Raises self.Error if s cannot be found.
        """
        for i, line in enumerate(self.lines):
            if s in line:
                return i
        else:
            raise self.Error(f"Cannot find `{s}` in lines")

    def get_input_str(self) -> str:
        """String with the ONCVPSP input file."""
        try:
            # oncvpsp 3.2.3
            i = self.find_string("<INPUT>")
            j = self.find_string("</INPUT>")
            return "\n".join(self.lines[i+1:j]) + "\n"
        except self.Error:
            # oncvpsp => 4
            i = self.find_string("Reference configufation results")
            return "\n".join(self.lines[:i])

    def get_psp8_str(self) -> Union[str, None]:
        """
        Return string with the pseudopotential data in psp8 format.
        Return None if field is not present.
        """
        start, stop = None, None
        for i, line in enumerate(self.lines):
            if 'Begin PSPCODE8' in line: start = i
            if start is not None and 'END_PSP' in line:
                stop = i
                break

        if start is None and stop is None: return None
        ps_data = "\n".join(self.lines[start+1:stop])

        if "<INPUT>" not in ps_data:
            # oncvpsp <= 3.2.2 --> Append the input to ps_data (note XML markers)
            # oncvpsp >= 3.2.3 --> Input is already there
            ps_data += "\n\n<INPUT>\n" + self.get_input_str() + "</INPUT>\n"

        return ps_data

    def get_upf_str(self) -> Union[str, None]:
        """
        Return string with the pseudopotential data in upf format.
        Return None if field is not present.
        """
        start, stop = None, None
        for i, line in enumerate(self.lines):
            if "Begin PSP_UPF" in line: start = i
            if start is not None and 'END_PSP' in line:
                stop = i
                break

        if start is None and stop is None: return None
        return "\n".join(self.lines[start+1:stop])

    def get_plotter(self): # -> Union[OncvPlotter, None]:
        """
        Return an instance of OncvPlotter or None
        """
        from abipy.ppcodes.oncv_plotter import OncvPlotter
        try:
            return OncvPlotter(self)
        except Exception as exc:
            print(exc)
            #raise
            return None

    def _grep(self, tag: str, beg: int = 0) -> GrepResults:
        """
        Finds the first field in the file with the specified tag.
        `beg` gives the initial position in the file.
        """
        data, stop, intag = [], None, -1

        if beg >= len(self.lines):
            raise ValueError(f"beg {beg} > len(lines) ({len(self.lines)})")

        for i, l in enumerate(self.lines[beg:]):
            l = l.lstrip()
            if l.startswith(tag):
                if intag == -1:
                    intag = beg + i
                data.append([float(c) for c in l.split()[1:]])
            else:
                # Exit because we know there's only one section starting with 'tag'
                if intag != -1:
                    stop = beg + i
                    break

        if not data:
            return GrepResults(data=None, start=intag, stop=stop)
        else:
            return GrepResults(data=np.array(data), start=intag, stop=stop)

    def gnuplot(self) -> None:
        """
        Plot the results with gnuplot.
        Based on the `replot.sh` script provided by the oncvpsp code.
        """
        outfile = self.filepath
        base = os.path.basename(outfile)
        gnufile = base + ".scr"
        plotfile = base + ".plot"
        temp = base + ".tmp"

        workdir = tempfile.mkdtemp()
        print(f"Working in: {workdir}")

        from monty.os import cd
        from subprocess import check_call
        with cd(workdir):
            check_call("awk 'BEGIN{out=0};/GNUSCRIPT/{out=0}; {if(out == 1) {print}}; \
                                /DATA FOR PLOTTING/{out=1}' %s > %s" % (outfile, plotfile), shell=True)
            check_call("awk 'BEGIN{out=0};/END_GNU/{out=0}; {if(out == 1) {print}}; \
                                /GNUSCRIPT/{out=1}' %s > %s" % (outfile, temp), shell=True)
            check_call('sed -e 1,1000s/t1/"%s"/ %s > %s' % (plotfile, temp, gnufile), shell=True)

            try:
                check_call(["gnuplot", gnufile])
            except KeyboardInterrupt:
                print("Received KeyboardInterrupt")

        os.rmdir(workdir)

    def get_atomic_levels_df(self) -> pd.DataFrame:
        """
        Return pandas dataframe with the atomic levels. Columns: (n, l, k, eig, occ, is_valence)
        """
        d_list = []
        from dataclasses import asdict
        for level in self.atomic_levels:
            data = asdict(level)
            d = data.pop("nlk").get_dict4pandas()
            d.update(**data)
            d_list.append(d)

        return pd.DataFrame(d_list)

    def get_peaks_df(self) -> pd.DataFrame:
        """
        Return pandas dataframe with the position of the last peak.
        """
        d_list = []

        def _push(typ, nkl, wf) -> None:
            peaks = wf.get_peaks()
            d = nlk.get_dict4pandas()
            d.update({"type": typ, "name": str(nlk), "last_peak_au": peaks.xs[-1]})
            d_list.append(d)

        ae_wfs, ps_wfs = self.radial_wfs.ae, self.radial_wfs.ps
        for nlk, ae_wf in ae_wfs.items():
            _push("AE", nlk, ae_wf)
            _push("PS", nlk, ps_wfs[nlk])

        return pd.DataFrame(d_list)



#!/usr/bin/env python
r"""
GW with scissors operator
=========================

This example shows how to generate an energy-dependent scissors operator
by fitting the GW corrections as function of the KS eigenvalues.
We then use the scissors operator to correct the KS band structure
computed on a high symmetry k-path. Finally, the LDA and the QPState band
structure are plotted with matplotlib.
"""
import abipy.data as abidata
from abipy.abilab import abiopen, ElectronBandsPlotter

# Get the QP results from the SIGRES.nc database.
sigma_file = abiopen(abidata.ref_file("si_g0w0ppm_nband30_SIGRES.nc"))

# Let's have a look at the QP correction as function of the KS energy.
# Don't shift KS eigenvalues to have zero energy at the Fermi energy.
# because we need the absolute values for the fit.
# The qpeme0(e0) curve consists of two branches:
#   the one in the [-6, 5.7] eV interval associated to valence states
#   and the one in the [6.1, 15] eV interval associated to conduction bands.
# We will fit these results with 2 functions defined in these two domains.
sigma_file.plot_qps_vs_e0(e0=None)
sigma_file.plot_qps_vs_e0()

qplist_spin = sigma_file.qplist_spin

# Define the two domains and construct the scissors operator
domains = [[-10, 6.1], [6.1, 18]]
scissors = qplist_spin[0].build_scissors(domains, bounds=None)

# Read the KS band energies computed on the k-path
ks_bands = abiopen(abidata.ref_file("si_nscf_GSR.nc")).ebands

# Read the KS band energies computed on the Monkhorst-Pack (MP) mesh
# and compute the DOS with the Gaussian method
ks_mpbands = abiopen(abidata.ref_file("si_scf_GSR.nc")).ebands
ks_edos = ks_mpbands.get_edos()

# Apply the scissors operator first on the KS band structure
# along the k-path then on the energies computed with the MP mesh.
qp_bands = ks_bands.apply_scissors(scissors)

qp_mpbands = ks_mpbands.apply_scissors(scissors)

# Compute the DOS with the modified QPState energies.
qp_edos = qp_mpbands.get_edos()

# Plot the LDA and the QPState band structure with matplotlib.
plotter = ElectronBandsPlotter()

plotter.add_ebands("LDA", ks_bands, edos=ks_edos)
plotter.add_ebands("LDA+scissors(e)", qp_bands, edos=qp_edos)

# By default, the two band energies are shifted wrt to *their* fermi level.
# Use e=0 if you don't want to shift the eigenvalus
# so that it's possible to visualize the QP corrections.
plotter.combiplot(title="Silicon band structure")

plotter.gridplot(title="Silicon band structure")
sigma_file.close()



#!/usr/bin/env python
r"""
Relaxation of Al with different k-meshes and tsmear
===================================================

In this example, we employ the relaxation algorithms implemented in Abinit (``ionmov`` and ``optcell``)
to find the equilibrium configuration of GaN (atomic positions and lattice vectors).
The relaxation is done with different k-meshes to monitor the convergence of the results.
You will observe a change of the equilibrium parameters with respect to the k-point mesh.

Note the we are using pseudopotentials generated with the GGA which tends to
overestimate the lattice parameters and ecut is way too low.
If you replace GGA with LDA, you will observe that LDA tends to underestimate the parameters.
"""

import sys
import os

import abipy.abilab as abilab
import abipy.flowtk as flowtk
import abipy.data as abidata


def relax_input(tsmear, nksmall):
    """
    Crystalline aluminum: optimization of the lattice parameter
    at fixed number of k-points and broadening.
    Similar to tbase4_1.in with minor modifications.
    """
    #structure = abilab.Structure.fcc()
    inp = abilab.AbinitInput(structure=abidata.ucells.structure_from_ucell("Al"),
                             pseudos=abidata.pseudos("13al.981214.fhi"))

    # Define k-point sampling.
    # nshiftk and shift are automatically selected from the lattice and the number of divisions
    # for the smallest direction. nksmall 2 e.g. will automatically select
    #   ngkpt 2 2 2
    #   nshiftk 4
    #   shiftk
    #       0.5 0.5 0.5
    #       0.5 0.0 0.0
    #       0.0 0.5 0.0
    #       0.0 0.0 0.5
    inp.set_autokmesh(nksmall=nksmall)

    inp.set_vars(
        ecut=6,
        occopt=4,
        tsmear=tsmear,
        toldfe=1e-6,
        nstep=10,
        optcell=1,    # Optimization of the lattice parameters
        ionmov=2,
        ntime=10,
        dilatmx=1.05,
        ecutsm=0.5,
        ixc=1,
    )

    return inp


def build_flow(options):
    """
    Build and return a flow performing structural relaxations with different k-point samplings.
    """
    # Set working directory (default is the name of the script with '.py' removed and "run_" replaced by "flow_")
    if not options.workdir:
        options.workdir = os.path.basename(sys.argv[0]).replace(".py", "").replace("run_", "flow_")

    # Let generate multiple inputs for different (tsmear, nksmall)
    # Product computes the Cartesian product of input iterables.
    # It's equivalent to nested for-loops
    tsmear_list = (0.01, 0.02, 0.03, 0.04)
    nksmall_list = (2, 4, 6)

    from itertools import product
    inputs = [relax_input(tsmear, nksmall) for tsmear, nksmall in product(tsmear_list, nksmall_list)]

    # Build flow form inputs.
    # As the calculations are independent, we can use Flow.from_inputs
    # Note the Flow.from_inputs is a simplified interface that, by default, builds tasks
    # for Ground-state calculation (GsTask).
    # Here we are performing a structural relaxation so we have to specify the task class explicitly.
    # AbiPy will use this piece of information to handle the restart of the RelaxTask that differs
    # from the one provided by GsTask.

    return flowtk.Flow.from_inputs(options.workdir, inputs=inputs, task_class=flowtk.RelaxTask)


# This block generates the thumbnails in the AbiPy gallery.
# You can safely REMOVE this part if you are using this script for production runs.
if os.getenv("READTHEDOCS", False):
    __name__ = None
    import tempfile
    options = flowtk.build_flow_main_parser().parse_args(["-w", tempfile.mkdtemp()])
    build_flow(options).graphviz_imshow()


@flowtk.flow_main
def main(options):
    """
    This is our main function that will be invoked by the script.
    flow_main is a decorator implementing the command line interface.
    Command line args are stored in `options`.
    """
    return build_flow(options)


if __name__ == "__main__":
    sys.exit(main())

############################################################################
# Run the script with:
#
#     run_relax_vs_kpts_tsmear.py -s
#
# then use:
#
#     abirun.py flow_relax_vs_kpts_tsmear hist -p
#
# to print (and plot) the structural relaxation for all the tasks:
#
# .. image:: https://github.com/abinit/abipy_assets/blob/master/run_relax_vs_kpts_tsmear_hist.png?raw=true
#    :alt: Commbiplot of the HIST files for the different (nkpt, tsmear) params
#
# To analyze the convergence of the relaxed lattice parameters, use:
#
# .. code-block:: bash
#
#	abirun.py flow_relax_vs_kpts_tsmear robot GSR
#
# to create a GSR robot for all the tasks in the flow and open an ipyton shell.
#
# Then, inside ipython, type:
#
# .. code-block:: ipython
#
#	 In [1]: %matplotlib
#	 In [2]: df = robot.get_dataframe()
#	 # Let's do some math with pandas to retrieve the Abinit acell from the a lattice parameter given in Ang.
#	 In [3]: import math
#	 In [4]: from abipy import abilab
#	 In [5]: df["acell"] = df["a"] * math.sqrt(2) * abilab.units.ang_to_bohr

#        In [7]: df["acell"]
#        Out[7]:
#        flow_relax_vs_kpts_tsmear/w0/t0/outdata/out_GSR.nc     7.558770
#        flow_relax_vs_kpts_tsmear/w0/t1/outdata/out_GSR.nc     7.505486
#        flow_relax_vs_kpts_tsmear/w0/t2/outdata/out_GSR.nc     7.496158
#        flow_relax_vs_kpts_tsmear/w0/t3/outdata/out_GSR.nc     7.558770
#        flow_relax_vs_kpts_tsmear/w0/t4/outdata/out_GSR.nc     7.505643
#        flow_relax_vs_kpts_tsmear/w0/t5/outdata/out_GSR.nc     7.495546
#        flow_relax_vs_kpts_tsmear/w0/t6/outdata/out_GSR.nc     7.558770
#        flow_relax_vs_kpts_tsmear/w0/t7/outdata/out_GSR.nc     7.501756
#        flow_relax_vs_kpts_tsmear/w0/t8/outdata/out_GSR.nc     7.496770
#        flow_relax_vs_kpts_tsmear/w0/t9/outdata/out_GSR.nc     7.558771
#        flow_relax_vs_kpts_tsmear/w0/t10/outdata/out_GSR.nc    7.504096
#        flow_relax_vs_kpts_tsmear/w0/t11/outdata/out_GSR.nc    7.499134
#        Name: acell, dtype: float64
#
#        # to plot the optimized acell vs nkpt for the different values of tsmear, use:
#	 In [6]: robot.plot_xy_with_hue(df, "nkpt", "acell", hue="tsmear")
#
# .. image:: https://github.com/abinit/abipy_assets/blob/master/run_relax_vs_kpts_tsmear.png?raw=true
#    :alt: optimized acell as function of nkpt and tsmear



#!/usr/bin/env python
r"""
Conductivity in metals
======================

Flow to compute conductivity in metals
"""

import os
import sys
import abipy.data as abidata
import abipy.abilab as abilab

from abipy import flowtk
from abipy.abio.factories import conduc_kerange_from_inputs


def make_scf_input(structure, pseudos, ngkpt=(2,2,2), shiftk=(0,0,0),
                   **variables):
    """Build and return SCF input given the structure and pseudopotentials"""

    scf_inp = abilab.AbinitInput(structure, pseudos=pseudos)

    # Global variables
    scf_inp.set_vars(**variables)

    # Dataset 1 (GS run)
    scf_inp.set_kmesh(ngkpt=ngkpt, shiftk=shiftk)
    #scf_inp.set_vars(toldfe=1e-10)

    return scf_inp


def make_nscf_input(structure, pseudos, ngkpt=(2,2,2), shiftk=(0,0,0),
                    **variables):
    """Build and return NSCF input given the structure and pseudopotentials"""
    scf_inp = abilab.AbinitInput(structure, pseudos=pseudos)

    # Global variables
    scf_inp.set_vars(**variables)

    # Dataset 1 (GS run)
    scf_inp.set_kmesh(ngkpt=ngkpt, shiftk=shiftk)
    scf_inp.set_vars(iscf=-2)

    return scf_inp


def build_flow(options):
    # Working directory (default is the name of the script with '.py' removed and "run_" replaced by "flow_")
    if not options.workdir:
        options.workdir = os.path.basename(__file__).replace(".py", "").replace("run_", "flow_")

    # Get structure and pseudos from the abipy database
    structure = abidata.structure_from_ucell("Al")
    pseudos = abidata.pseudos("13al.pspnc")

    # Variables
    variables = dict(
            ecut=20,
            tsmear=0.05,
            nband=12,
            nbdbuf=2,
            occopt=3,
            iomode=1,
            nstep=20
    )

    ngkpt = [4, 4, 4]
    ngkpt_fine = [8, 8, 8]
    shiftk = [0.0, 0.0, 0.0]
    ngqpt = [2, 2, 2]
    tmesh = [0, 30, 11] # Conductivity at temp from 0K to 300K by increment of 30
    boxcutmin = 1.1
    mixprec = 1

    # Kerange Variables
    nbr_proc = 4
    ngqpt_fine = [16, 16, 16] # The sigma_ngkpt grid must be divisible by the qpt grid
    sigma_ngkpt = [16, 16, 16]
    einterp = [1, 5, 0, 0] # Star functions Interpolation
    sigma_erange = [-0.3, -0.3, "eV"] # Negative value for metals

    flow = flowtk.Flow(workdir=options.workdir)

    # Create inputs Object
    scf_input = make_scf_input(structure, pseudos,
                               tolvrs=1e-12,
                               ngkpt=ngkpt,
                               shiftk=shiftk,
                               **variables)

    nscf_input = make_nscf_input(structure, pseudos,
                                 tolwfr=1e-18,
                                 ngkpt=ngkpt_fine,
                                 shiftk=shiftk,
                                 **variables)

    # Create Work Object
    # Work 0:  SCF run
    gs_work = flowtk.Work()
    gs_work.register_scf_task(scf_input)
    flow.register_work(gs_work)

    # Work 1: Compute DDB et DVDB
    ph_work = flowtk.PhononWork.from_scf_task(gs_work[0],
                                              qpoints=ngqpt, is_ngqpt=True,
                                              tolerance={"tolvrs": 1e-8})
    flow.register_work(ph_work)

    # Work 2: Conductivity with Kerange
    multi = conduc_kerange_from_inputs(scf_input=scf_input,
                                       nscf_input=nscf_input,
                                       tmesh=tmesh,
                                       ddb_ngqpt=ngqpt,
                                       eph_ngqpt_fine=ngqpt_fine,
                                       sigma_ngkpt=sigma_ngkpt,
                                       sigma_erange=sigma_erange,
                                       einterp=einterp,
                                       boxcutmin=boxcutmin, # 1.1 is the default value of the function
                                       mixprec=mixprec # 1 is the default value of the function
                                       )

    # Here we can change multi to change the variable of a particular dataset

    conduc_work = flowtk.ConducWork.from_phwork(phwork=ph_work, # Linking the DDB and DVDB via a |PhononWork|
                                                multi=multi, # The multidataset object
                                                nbr_proc=nbr_proc, # Needed to parallelize the calculation
                                                flow=flow,
                                                with_kerange=True, # Using Kerange
                                                omp_nbr_thread=1) # 1 is the default value of the function
    # If you already have the DDB and DVDB, use from_filepath(ddb_path, dvdb_path, multi, ...) instead of from_phwork

    flow.register_work(conduc_work)

    return flow.allocate(use_smartio=True)


# This block generates the thumbnails in the Abipy gallery.
# You can safely REMOVE this part if you are using this script for production runs.
if os.getenv("READTHEDOCS", False):
    __name__ = None
    import tempfile
    options = flowtk.build_flow_main_parser().parse_args(["-w", tempfile.mkdtemp()])
    build_flow(options).graphviz_imshow()


@flowtk.flow_main
def main(options):
    """
    This is our main function that will be invoked by the script.
    flow_main is a decorator implementing the command line interface.
    Command line args are stored in `options`.
    """
    return build_flow(options)


if __name__ == "__main__":
    sys.exit(main())

############################################
